import json
import curses
import textwrap
import functools
from typing import Optional
from base64 import b16encode  # for ping/pong
from os import urandom

import humanize

from twisted.internet.task import deferLater
from twisted.internet.defer import ensureDeferred, race
from twisted.internet.stdio import StandardIO
from twisted.protocols.basic import LineReceiver

from wormhole.errors import LonelyError

import attr

from .observer import Next, When
from ._proto import wormhole_from_config, create_fowl, fowld_output_to_json
from .messages import (
    Welcome,
    CodeAllocated,
    PeerConnected,
    WormholeClosed,
    AllocateCode,
    SetCode,
    LocalListener,
    RemoteListener,
    Listening,
    RemoteListeningSucceeded,
    BytesIn,
    BytesOut,
    IncomingConnection,
    IncomingDone,
    IncomingLost,
    OutgoingConnection,
    OutgoingDone,
    WormholeError,
    GrantPermission,
    Ping,
    Pong,
)


@attr.frozen
class Connection:
    i: int = 0
    o: int = 0
    listener_id: str = None  # Maybe<str>; Nothing if incoming connection


@attr.frozen
class State:
    code: Optional[str] = None
    connected: bool = False
    verifier: Optional[str] = None
    listeners: list = attr.Factory(list)
    remote_listeners: list = attr.Factory(list)
    connections: dict[int, Connection] = attr.Factory(dict)

    @property
    def pretty_verifier(self):
        # space-ify this, for easier reading
        return " ".join(
            self.verifier[a:a+4]
            for a in range(0, len(self.verifier), 4)
        )


async def frontend_tui(reactor, config):
    print(f"Connecting: {config.relay_url}")

    @functools.singledispatch
    def output_message(msg):
        print(f"\b\b\b\bunhandled output: {msg}")

    @output_message.register(Pong)
    def _(msg):
        print(f"\b\b\b\b  <- Pong({b16encode(msg.ping_id).decode('utf8')}): {msg.time_of_flight}s\n>>> ", end="")

    @output_message.register(WormholeError)
    def _(msg):
        print(f"\b\b\b\bERROR: {msg.message}")

    @output_message.register(Listening)
    def _(msg):
        print(f"\b\b\b\bListening: {msg.listen}")
        replace_state(attr.evolve(state[0], listeners=state[0].listeners + [msg]))

    @output_message.register(RemoteListeningSucceeded)
    def _(msg):
        print(f"\b\b\b\bRemote side is listening: {msg.listen}")
        replace_state(attr.evolve(state[0], remote_listeners=state[0].remote_listeners + [msg]))

    @output_message.register(IncomingConnection)
    def _(msg):
        conn = state[0].connections
        conn[msg.id] = Connection(0, 0, msg.listener_id)
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(IncomingDone)
    def _(msg):
        print(f"\b\b\b\bClosed: {msg.id}")
        conn = state[0].connections
        del conn[msg.id]
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(IncomingLost)
    def _(msg):
        print(f"\b\b\b\bLost: {msg.id}: {msg.reason}")
        conn = state[0].connections
        del conn[msg.id]
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(OutgoingConnection)
    def _(msg):
        conn = state[0].connections
        conn[msg.id] = Connection(0, 0, msg.listener_id)
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(OutgoingDone)
    def _(msg):
        conn = state[0].connections
        print(f"\b\b\b\bClosed: {msg.id} from {conn[msg.id].listener_id}")
        del conn[msg.id]
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(BytesIn)
    def _(msg):
        conn = state[0].connections
        conn[msg.id] = attr.evolve(conn[msg.id], i=conn[msg.id].i + msg.bytes)
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(BytesOut)
    def _(msg):
        conn = state[0].connections
        conn[msg.id] = attr.evolve(conn[msg.id], o=conn[msg.id].o + msg.bytes)
        replace_state(attr.evolve(state[0], connections=conn))

    @output_message.register(WormholeClosed)
    def _(msg):
        print(f"Closed({msg.result})...", end="", flush=True)
        # close our standard input, so the human can't give us more
        # commands -- because we're done
        command_reader.transport.loseConnection()

    @output_message.register(Welcome)
    def _(msg):
        print("\b\b\b\b", end="")
        print("Connected.")
        if "motd" in msg.welcome:
            print(textwrap.fill(msg.welcome["motd"].strip(), 80, initial_indent="    ", subsequent_indent="    "))
        print(">>> ", end="", flush=True)

    start_time = reactor.seconds()
    if config.output_debug_messages:
        def output_wrapper(msg):
            try:
                js = fowld_output_to_json(msg)
                # don't leak our absolute time, more convenient anyway
                js["timestamp"] = reactor.seconds() - start_time
                config.output_debug_messages.write(
                    json.dumps(js) + "\n"
                )
            except Exception as e:
                print(e)
            return output_message(msg)
    else:
        output_wrapper = output_message

    fowl_wh = await create_fowl(config, output_wrapper)

    # make into IService?
    fowl_wh.start()

    state = [State()]

    def replace_state(new_state):
        ##print("replace with", new_state)
        ##print(state[0])
        old = state[0]
        new_output = "\b\b\b\b"
        if new_state.connected and not old.connected:
            new_output += "Connected to peer!\n"
        if new_state.code and old.code is None:
            new_output += "Code: {}\n".format(new_state.code)
        if new_state.verifier and old.verifier is None:
            new_output += "Verifier: {}\n".format(new_state.pretty_verifier)
        for conid, conn in new_state.connections.items():
            b = conn.i + conn.o
            if b:
                new_output += f"{conid}: {humanize.naturalsize(b)}\n"
        if new_output:
            print(f"{new_output}>>> ", end="", flush=True)
        state[0] = new_state

    @output_message.register(CodeAllocated)
    def _(msg):
        replace_state(attr.evolve(state[0], code=msg.code))

    @output_message.register(PeerConnected)
    def _(msg):
        replace_state(attr.evolve(state[0], connected=True, verifier=msg.verifier))

    create_stdio = config.create_stdio or StandardIO
    command_reader = CommandReader(reactor)
    create_stdio(command_reader)

    print(">>> ", end="", flush=True)
    while True:
        what, result = await race([
            ensureDeferred(command_reader.next_command()),
            ensureDeferred(command_reader.when_closed()),
        ])
        if what == 0:
            cmd_line = result
            if cmd_line.strip():
                cmd = cmd_line.decode("utf8").split()
                cmd_name = cmd[0]
                try:
                    cmd_fn = commands[cmd_name]
                except KeyError:
                    if cmd_name.strip().lower() == "quit" or cmd_name.strip().lower() == "q":
                        break
                    print(f'No such command "{cmd_name}"')
                    print("Commands: {}".format(" ".join(commands.keys())))
                    print("Ctrl-d to quit")
                    print(">>> ", end="", flush=True)
                    continue
                # XXX should be passing "high level" FowlWormhole thing, not Wormhole direct
                await cmd_fn(reactor, fowl_wh, state[0], *cmd[1:])
            else:
                print(">>> ", end="", flush=True)
        elif what == 1:
            break

    print("\nClosing mailbox...", end="", flush=True)
    try:
        await fowl_wh.stop()
    except LonelyError:
        pass
    print("done.")


async def _cmd_help(reactor, wh, state, *args):
    """
    Some helpful words
    """
    funs = dict()
    for name, fn in commands.items():
        try:
            funs[fn].append(name)
        except KeyError:
            funs[fn] = [name]
    for fn, aliases in funs.items():
        name = sorted(aliases)[-1]
        rest = " ".join(sorted(aliases)[:-1])
        helptext = textwrap.dedent(fn.__doc__)
        if helptext:
            print(f"{name} ({rest})")
            print(textwrap.fill(helptext.strip(), 80, initial_indent="    ", subsequent_indent="    "))
            print()


async def _cmd_invite(reactor, wh, state, *args):
    """
    Allocate a code (to give to a peer)
    """
    if args:
        print("No arguments allowed")
        return
    # XXX fixme no private usage
    if state.code is not None:
        print(f"Existing code: {state.code}")
    else:
        wh.command(AllocateCode())


async def _cmd_accept(reactor, wh, state, *args):
    """
    Consume an already-allocated code (from a peer)
    """
    if len(args) != 1:
        print('Require a secret code (e.g. from "invite" on the other side)')
        return
    if state.code is not None:
        print(f"Existing code: {state.code}")
    else:
        wh.command(SetCode(args[0]))


async def _cmd_listen_local(reactor, wh, state, *args):
    """
    Listen locally on the given port; connect to the remote side on
    the same port (or a custom one if two ports are passed)
    """
    try:
        port = int(args[0])
    except (ValueError, IndexError):
        print("Requires a TCP port, as an integer.")
        print("We will listen on this TCP port on localhost, and connect the same")
        print("localhost port on the far side. Optionally, a second port may be")
        print("specified to use a different far-side port")
        return

    if len(args) > 1:
        try:
            remote_port = int(args[1])
        except ValueError:
            print(f"Not port-number: {args[1]}")
            return
    else:
        remote_port = port

    wh.command(
        GrantPermission(
            listen=[port],
            connect=[],
        )
    )
    wh.command(
        LocalListener(
            listen=f"tcp:{port}:interface=localhost",
            connect=f"tcp:localhost:{remote_port}",
        )
    )


async def _cmd_listen_remote(reactor, wh, state, *args):
    """
    Listen on the remote side on the given port; connect back to this
    side on the same port (or a custom one if two ports are passed)
    """
    try:
        remote_port = int(args[0])
    except (ValueError, IndexError):
        print("Requires a TCP port, as an integer.")
        print("We will listen on this TCP port on the remote side and connect to the same")
        print("localhost port on this side. Optionally, a second port may be specified")
        print("to use a different local port")
        return

    if len(args) > 1:
        try:
            local_port = int(args[1])
        except ValueError:
            print(f"Not port-number: {args[1]}")
            return
    else:
        local_port = remote_port

    wh.command(
        GrantPermission(
            listen=[],
            connect=[local_port],
        )
    )
    wh.command(
        RemoteListener(
            listen=f"tcp:{remote_port}:interface=localhost",
            connect=f"tcp:localhost:{local_port}",
        )
    )


async def _cmd_allow(reactor, wh, state, *args):
    """
    Allow an incoming connection on a particular port
    """
    try:
        local_port = int(args[0])
    except (ValueError, IndexError):
        print("Requires a TCP port, as an integer.")
        print("If the other side tries to connect via this port, we will allow it")
        return
    wh.command(
        GrantPermission(
            listen=[],
            connect=[local_port],
        )
    )


async def _cmd_allow_listen(reactor, wh, state, *args):
    """
    Allow remote side to listen on a given TCP port.
    """
    try:
        local_port = int(args[0])
    except (ValueError, IndexError):
        print("Requires a TCP port, as an integer.")
        print("We will allow the other side to listen on this TCP port")
        return
    wh.command(
        GrantPermission(
            listen=[local_port],
            connect=[],
        )
    )


async def _cmd_ping(reactor, wh, state, *args):
    """
    Send a ping (through the Mailbox Server)
    """
    ping_id = None
    if len(args) != 0:
        raise Exception("No argument accepted to ping")
    ping_id = urandom(4)
    print(f"  -> Ping({b16encode(ping_id).decode('utf8')})\n>>> ", end="")
    wh.command(Ping(ping_id))


async def _cmd_status(reactor, wh, state, *args):
    print("status")
    peer = "disconnected"
    if state.connected:
        peer = "yes"
    if state.verifier:
        peer += f" verifier={state.verifier}"
    else:
        print(f"  code: {state.code}")
    print(f"  peer: {peer}")
    if state.listeners:
        print("  listeners:")
        for listener in state.listeners:
            print(f"    {listener.listener_id.lower()}: {listener.listen} -> {listener.connect}")
    if state.remote_listeners:
        print("  remote listeners:")
        for listener in state.remote_listeners:
            print(f"    {listener.listener_id}: {listener.connect} <- {listener.listen}")
    if state.connections:
        print("  connections:")
        for conn_id, conn in state.connections.items():
            listener = ""
            if conn.listener_id is not None:
                listener = f"  (via {conn.listener_id})"
            print(f"    {conn_id}: {conn.i} bytes in / {conn.o} bytes out{listener}")
    print(">>> ", end="", flush=True)


class CommandReader(LineReceiver):
    """
    Wait for incoming commands from the user
    """
    delimiter = b"\n"
    _next_command = None
    _closed = None
    reactor = None

    def __init__(self, reactor):
        super().__init__()
        self.reactor = reactor
        self._next_command = Next()
        self._closed = When()

    async def next_command(self):
        return await self._next_command.next_item()

    async def when_closed(self):
        return await self._closed.when_triggered()

    def connectionMade(self):
        pass

    def lineReceived(self, line):
        self._next_command.trigger(self.reactor, line)

    def connectionLost(self, why):
        # beware: triggering with "why" goes to errback chain, not
        # what we usually want
        self._closed.trigger(self.reactor, None)


## okay, well that is entertaining, but want something quicker to
## facilitate refactoring, so "REPL-style" it'll be for now?
## ...but would be great to expand that to have "status" style
## info "above" the repl stuff, e.g. bandwidth etc
async def curses_frontend_tui(reactor, config):
    w = await wormhole_from_config(config)

    # XXX does "curses.wrapper" work with async functions?
    ##curses.wrapper(partial(_tui, config, w))

    try:
        stdscr = curses.initscr()
        curses.noecho()
        curses.cbreak()
        curses.curs_set(0)
        stdscr.keypad(True)
        stdscr.nodelay(True)  # non-blocking getch()

        await _tui(reactor, config, w, stdscr)

    finally:
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()
        curses.endwin()


async def sleep(reactor, delay):
    await deferLater(reactor, delay, lambda: None)


async def _tui(reactor, config, w, stdscr):
    stdscr.clear()

    d0 = w.get_welcome()
    d1 = w.get_code()
    d2 = w.get_versions()

    stdscr.addstr(0, 1, "welcome:")
    stdscr.addstr(1, 4, "code:")
    stdscr.addstr(2, 0, "versions:")
    stdscr.refresh()

    def got_welcome(js):
        stdscr.addstr(0, 1, f"welcome: {js}")
        stdscr.refresh()
        w.allocate_code()
        return js
    d0.addCallback(got_welcome)

    def got_code(code):
        stdscr.addstr(1, 4, f"code: {code}")
        stdscr.refresh()
        return code
    d1.addCallback(got_code)

    def got_versions(versions):
        w.dilate()
        stdscr.addstr(2, 0, f"versions: {versions}")
        stdscr.refresh()
        return versions
    d2.addCallback(got_versions)

    while True:
        # NOTE: it's vital to "wait" or something on the reactor,
        # otherwise we're madly busy-looping in curses (only!) and
        # nothing else can happen
        await sleep(reactor, 0.1)
        k = stdscr.getch()
        if k != curses.ERR:
            if k == ord('q') or k == ord('Q') or k == 27:
                break

    try:
        await w.close()
    except LonelyError:
        pass


commands = {
    "invite": _cmd_invite,
    "i": _cmd_invite,

    "accept": _cmd_accept,
    "a": _cmd_accept,

    "local": _cmd_listen_local,
    "remote": _cmd_listen_remote,

    "allow": _cmd_allow,
    "allow-listen": _cmd_allow_listen,

    "ping": _cmd_ping,
    "p": _cmd_ping,

    "status": _cmd_status,
    "st": _cmd_status,
    "s": _cmd_status,

    "help": _cmd_help,
    "h":_cmd_help,
    "?": _cmd_help,
}




import ipaddress

from attr import frozen
from zope.interface import Interface, implementer
from twisted.internet.endpoints import (
    TCP4ServerEndpoint,
    TCP6ServerEndpoint,
    TCP4ClientEndpoint,
    TCP6ClientEndpoint,
)


class IClientListenPolicy(Interface):
    """
    A way to ask which endpoints are acceptable to listen upon for a client.
    """

    def can_listen(self, endpoint) -> bool:
        """
        :returns: True if the given IStreamServerEndpoint is acceptable to
            this policy
        """


class IClientConnectPolicy(Interface):
    """
    Ask what endpoints are acceptable to connect on
    """

    def can_connect(self, endpoint) -> bool:
        """
        :returns: True if the given IStreamClientEndpoint is acceptable to
            this policy
        """


# XXX if i'm offline, "localhost" doesn't work (with ip_address()) -- when _does_ it work, and why?
# XXX with radios off entirely, i'm seeing "" (empty string) as the addr here
def is_localhost(addr: str) -> bool:
    if addr.strip() == "":
        return False
    if addr == "localhost":
        return True
    ip = ipaddress.ip_address(addr)
    return ip.is_loopback


@implementer(IClientListenPolicy)
class LocalhostAnyPortsListenPolicy:
    """
    Accepts any port as long as the interface is a local one (i.e. ::1
    or localhost or 127.0.0.1/8) according to the "ipaddress" library.
    """
    def can_listen(self, endpoint) -> bool:
        if isinstance(endpoint, (TCP6ServerEndpoint, TCP4ServerEndpoint)):
            return is_localhost(endpoint._interface)
        return False


@implementer(IClientListenPolicy)
@frozen
class LocalhostTcpPortsListenPolicy(LocalhostAnyPortsListenPolicy):
    # which ports we will accept
    ports: list[int]

    def can_listen(self, endpoint) -> bool:
        if super().can_listen(endpoint):
            # if we're here, parent has checked types too
            if endpoint._port in self.ports:
                return True
        return False


@implementer(IClientListenPolicy)
@frozen
class ArbitraryInterfaceTcpPortsListenPolicy:
    # interface, port pairs we accept
    listeners: list[tuple]

    def can_listen(self, endpoint) -> bool:
        if isinstance(endpoint, (TCP6ServerEndpoint, TCP4ServerEndpoint)):
            iface = endpoint._interface
            port = endpoint._port
            for allowed_iface, allowed_port in self.listeners:
                if is_localhost(allowed_iface) and is_localhost(endpoint._interface):
                    if port == allowed_port:
                        return True
                if iface == allowed_iface:
                    if port == allowed_port:
                        return True
        return False


@implementer(IClientListenPolicy)
class AnyListenPolicy:
    """
    Accepts any listener at all. DANGER.
    """
    def can_listen(self, endpoint) -> bool:
        return True


@implementer(IClientConnectPolicy)
class LocalhostAnyPortsConnectPolicy:
    """
    Accepts any port as long as the interface is a local one (i.e. ::1
    or localhost or 127.0.0.1/8) according to the "ipaddress" library.
    """
    def can_connect(self, endpoint) -> bool:
        if isinstance(endpoint, (TCP6ClientEndpoint, TCP4ClientEndpoint)):
            return is_localhost(endpoint._host)
        return False


@implementer(IClientListenPolicy)
@frozen
class ArbitraryAddressTcpConnectPolicy:
    # interface, port pairs we accept
    connecters: list[tuple]

    def can_connect(self, endpoint) -> bool:
        if isinstance(endpoint, (TCP6ClientEndpoint, TCP4ClientEndpoint)):
            addr, port = endpoint._host, endpoint._port
            for allowed_addr, allowed_port in self.connecters:
                if addr == allowed_addr:
                    if port == allowed_port:
                        return True
        return False


@implementer(IClientConnectPolicy)
class AnyConnectPolicy:
    """
    Accepts any connection at all. DANGER.
    """
    def can_connect(self, endpoint) -> bool:
        return True


@implementer(IClientConnectPolicy)
@frozen
class LocalhostTcpPortsConnectPolicy(LocalhostAnyPortsConnectPolicy):
    # which ports we will accept
    ports: list[int]

    def can_connect(self, endpoint) -> bool:
        if super().can_connect(endpoint):
            # if we're here, parent has checked types too
            if endpoint._port in self.ports:
                return True
        return False





import ipaddress
from hypothesis.strategies import ip_addresses, one_of, integers, lists
from hypothesis import given, assume

from twisted.internet.endpoints import TCP4ServerEndpoint, TCP6ServerEndpoint

from fowl.policy import LocalhostAnyPortsListenPolicy, LocalhostTcpPortsListenPolicy


def ip_address_to_listener(reactor, port, ipaddr):
    """
    Convert an ipaddress.IPv4Address or IPv6Address to an
    TCP4ServerEndpoint (or TCP6ServerEndpoint)
    """
    if isinstance(ipaddr, ipaddress.IPv4Address):
        return TCP4ServerEndpoint(reactor, port, interface=str(ipaddr))
    elif isinstance(ipaddr, ipaddress.IPv6Address):
        return TCP6ServerEndpoint(reactor, port, interface=str(ipaddr))
    raise ValueError(f"Unknown ipaddress: {ipaddr}")


@given(
    integers(1, 65536),  # any port we might care about
    one_of(
        ip_addresses(network="127.0.0.0/8"),
        ip_addresses(network="::1"),
        # these are "link-local" but I believe different meaning from "localhost"?
        # ip_addresses(network="fe80::/64"),
        # also what about "ipv4-mapped addresses that are actually localhost? possible? worth it?
    )
)
def test_policy_any_acceptable(reactor, port, ipaddr):
    """
    LocalhostAnyPortsPolicy allows all valid localhost style addresses
    we might like
    """
    endpoint = ip_address_to_listener(reactor, port, ipaddr)
    policy = LocalhostAnyPortsListenPolicy()
    assert policy.can_listen(endpoint) == True, "Listen on a localhost port"


@given(
    integers(1, 65536),  # any port we might care about
    ip_addresses(),
)
def test_policy_any_bad(reactor, port, ipaddr):
    """
    LocalhostAnyPortsPolicy disallows all IP addresses that are NOT
    localhost
    """
    assume(not ipaddress.ip_address(ipaddr).is_loopback)
    endpoint = ip_address_to_listener(reactor, port, ipaddr)
    policy = LocalhostAnyPortsListenPolicy()
    assert policy.can_listen(endpoint) == False, "Should only allow loopback addresses"


@given(
    integers(1, 65536),  # any port we might care about
    lists(integers(1, 65536)),  # allowed ports in our policy
    ip_addresses(),
)
def test_policy_specific_ports(reactor, port, allowed_ports, ipaddr):
    """
    LocalhostAnyPortsPolicy disallows all IP addresses that are NOT
    localhost
    """
    is_local = ipaddress.ip_address(ipaddr).is_loopback
    is_allowed = port in allowed_ports
    endpoint = ip_address_to_listener(reactor, port, ipaddr)
    expected_result = True if is_local and is_allowed else False

    policy = LocalhostTcpPortsListenPolicy(allowed_ports)
    assert policy.can_listen(endpoint) == expected_result, f"what port={endpoint._port} if={endpoint._interface} {expected_result} {port} {allowed_ports} {ipaddr} {policy.can_listen(endpoint)}"





#
# testing some ideas around storing + visualizing "bandwidth" data ... maybe any time-series data
#
# a while ago i played with "mutli-time-scale" things, e.g. say a
# 2-section one where the "oldest" section is 10x more time (or
# whatever). So e.g. 10 data-points in a row visually "collect" into a
# single roll-up (e.g. average .. or min/max) before proceeding
# through the second one.
#
# ...need a lot of history to "see" that though, so having a test-driver for this is good.
#
# ...for the rust thing, a full "fake fowl, that always opens a
# connection on ports you open" or something would be cool.
#




from typing import Optional
from attrs import frozen


class FowlOutputMessage:
    """
    An information message from fowld to the controller
    """


class FowlCommandMessage:
    """
    A command from the controller to fowld
    """


class FowlInternalControl:
    "A message from the state-machine to outside, basically?"
    pass


# if we had ADT / Union types, these would both be that -- is this as
# close as we can get in Python?


@frozen
class Welcome(FowlOutputMessage):
    """
    We have connected to the Mailbox Server and received the
    Welcome message.
    """
    url: str  # server address
    # open-ended information from the server
    welcome: dict


@frozen
class WormholeClosed(FowlOutputMessage):
    """
    The wormhole has been terminated.
    """
    result: str


@frozen
class AllocateCode(FowlCommandMessage):
    """
    Create a fresh code on the server
    """
    length: Optional[int] = None


@frozen
class SetCode(FowlCommandMessage):
    """
    Give a code we know to the server
    """
    code: str


@frozen
class CodeAllocated(FowlOutputMessage):
    """
    The secret wormhole code has been determined
    """
    code: str


@frozen
class PeerConnected(FowlOutputMessage):
    """
    We have evidence that the peer has connected
    """
    # hex-encoded 32-byte hash output (should match other peer)
    verifier: str
    versions: dict


@frozen
class GrantPermission(FowlCommandMessage):
    """
    Grant additional listen or connection privileges. Both are lists
    of valid ports between 1 and 65535 inclusive.
    """
    listen: list[int]
    connect: list[int]


@frozen
class DangerDisablePermissionCheck(FowlCommandMessage):
    """
    DANGER: allow listening or connecting to anything. Can be
    dangerous, you must know the implications.
    """


@frozen
class LocalListener(FowlCommandMessage):
    """
    We wish to open a local listener.
    """
    listen: str  # Twisted server-type endpoint string
    connect: str  # Twisted client-type endpoint string


@frozen
class RemoteListener(FowlCommandMessage):
    """
    We wish to open a listener on the peer.
    """
    listen: str  # Twisted server-type endpoint string
    connect: str  # Twisted client-type endpoint string


@frozen
class Ping(FowlCommandMessage):
    ping_id: int


@frozen
class Listening(FowlOutputMessage):
    """
    We have opened a local listener.

    Any connections to this listener will result in a subchannel and a
    connect on the other side (to "connected_endpoint"). This message
    may result from a LocalListener or a RemoteListener command. This
    message will always appear on the side that's actually listening.
    """
    listener_id: str  # random identifier
    listen: str  # Twisted server-type endpoint string
    connect: str  # Twisted client-type endpoint string


@frozen
class RemoteListeningFailed(FowlOutputMessage):
    """
    We have failed to open a listener on the remote side.
    """
    listen: str  # Twisted server-type endpoint string
    reason: str


@frozen
class RemoteListeningSucceeded(FowlOutputMessage):
    """
    The remote peer suceeded at fulfilling our listen request.
    """
    listener_id: str  # arbitrary identifier
    listen: str  # Twisted server-type endpoint string
    connect: str  # Twisted client-type endpoint string


@frozen
class RemoteConnectFailed(FowlOutputMessage):
    """
    Our peer could not connect
    """
    id: int
    reason: str


@frozen
class OutgoingConnection(FowlOutputMessage):
    """
    Something has connected to one of our listeners (and we are making
    an outgoing subchannel to the other peer).
    """
    id: int
    endpoint: str  # connection to here on far side
    listener_id: str  # which listener this is from


@frozen
class OutgoingLost(FowlOutputMessage):
    """
    We have lost one of our connections
    """
    id: int
    reason: str


@frozen
class OutgoingDone(FowlOutputMessage):
    """
    We have lost one of our connections
    """
    id: int


@frozen
class IncomingConnection(FowlOutputMessage):
    """
    The other side is requesting we open a connection
    """
    id: int
    endpoint: str
    listener_id: str


@frozen
class IncomingLost(FowlOutputMessage):
    """
    We have lost one of our connections
    """
    id: int
    reason: str


@frozen
class IncomingDone(FowlOutputMessage):
    """
    An incoming connection has ended successfully
    """
    id: int


@frozen
class BytesIn(FowlOutputMessage):
    id: int
    bytes: int


@frozen
class BytesOut(FowlOutputMessage):
    id: int
    bytes: int


@frozen
class WormholeError(FowlOutputMessage):
    message: str


@frozen
class PleaseCloseWormhole(FowlInternalControl):
    reason: str


@frozen
class Pong(FowlOutputMessage):
    ping_id: int
    time_of_flight: float


#XXX these aren't really used; why state-machine has them?
@frozen
class SendMessageToPeer(FowlOutputMessage):
    message: str


@frozen
class GotMessageFromPeer(FowlOutputMessage):
    message: str



from __future__ import print_function

import os
import sys
import json
import binascii
import textwrap
import functools
import struct
import signal
from base64 import b16encode
from typing import IO, Callable, TextIO
from functools import partial
from itertools import count

import click
import humanize
from attrs import frozen, define, asdict, Factory as AttrFactory

import msgpack
import automat
from twisted.internet import reactor
from twisted.internet.defer import Deferred, ensureDeferred, DeferredList, race, CancelledError
from twisted.internet.task import deferLater
from twisted.internet.endpoints import serverFromString, clientFromString
from twisted.internet.protocol import Factory, Protocol
from twisted.internet.error import ConnectionDone
from twisted.internet.stdio import StandardIO
from twisted.protocols.basic import LineReceiver
from zope.interface import directlyProvides
from wormhole.cli.public_relay import RENDEZVOUS_RELAY as PUBLIC_MAILBOX_URL
import wormhole.errors as wormhole_errors

from .observer import When
from .messages import (
    SetCode,
    AllocateCode,
    GrantPermission,
    DangerDisablePermissionCheck,
    RemoteListener,
    LocalListener,

    Welcome,
    Listening,
    RemoteListeningSucceeded,
    RemoteListeningFailed,
    RemoteConnectFailed,
    PeerConnected,
    CodeAllocated,
    BytesOut,
    BytesIn,
    IncomingConnection,
    IncomingDone,
    IncomingLost,
    OutgoingConnection,
    OutgoingDone,
    OutgoingLost,
    WormholeClosed,
    WormholeError,
    GotMessageFromPeer,

    Ping,
    Pong,

    FowlOutputMessage,
    FowlCommandMessage,

    PleaseCloseWormhole,
)
from .policy import IClientListenPolicy, IClientConnectPolicy, AnyConnectPolicy, AnyListenPolicy



APPID = u"meejah.ca/wormhole/forward"
WELL_KNOWN_MAILBOXES = {
    "default": PUBLIC_MAILBOX_URL,
    "local": "ws://localhost:4000/v1",
    "winden": "wss://mailbox.mw.leastauthority.com/v1",
    # Do YOU run a public mailbox service? Contact the project to
    # consider having it listed here
}
SUPPORTED_FEATURES = [
    "core-v1",  # the only version of the protocol so far
]


def _sequential_id():
    """
    Yield a stream of IDs, starting at 1
    """
    next_id = 0
    while True:
        next_id += 1
        yield next_id


allocate_connection_id = partial(next, _sequential_id())


#@frozen
@define  ## could be @frozen, but for "policy" ... hmmm
class _Config:
    """
    Represents a set of validated configuration
    """
    relay_url: str = PUBLIC_MAILBOX_URL
    code: str = None
    code_length: int = 2
    use_tor: bool = False
    appid: str = APPID
    debug_state: bool = False
    stdout: IO = sys.stdout
    stderr: IO = sys.stderr
    create_stdio: Callable = None  # returns a StandardIO work-alike, for testing
    debug_file: IO = None  # for state-machine transitions
    commands: list[FowlCommandMessage] = AttrFactory(list)
    connect_policy: IClientConnectPolicy = None
    listen_policy: IClientListenPolicy = None
    output_debug_messages: TextIO = None  # Option<Writable>


async def wormhole_from_config(reactor, config, wormhole_create=None):
    """
    Create a suitable wormhole for the given configuration.

    :returns DeferredWormhole: a wormhole API
    """
    if wormhole_create is None:
        from wormhole import create as wormhole_create

    tor = None
    if config.use_tor:
        tor = await get_tor(reactor)
        # XXX use a Message
        print(
            json.dumps({
                "kind": "tor",
                "version": tor.version,
            }),
            file=config.stdout,
            flush=True,
        )

    w = wormhole_create(
        config.appid or APPID,
        config.relay_url,
        reactor,
        tor=tor,
        timing=None,  # args.timing,
        _enable_dilate=True,
        versions={
            "fowl": {
                "features": SUPPORTED_FEATURES,
            }
        }
    )
    if config.debug_state:
        w.debug_set_trace("forward", file=config.stdout)
    return w


@frozen
class Connection:
    endpoint: str
    i: int = 0
    o: int = 0
    listener_id: str = "unknown"


async def frontend_accept_or_invite(reactor, config):

    connections = dict()
    when_explicitly_closed_d = Deferred()
    we_closed_d = Deferred()
    peer_connected = False

    @functools.singledispatch
    def output_message(msg):
        print(f"UNhandled output: {msg}")

    @output_message.register(WormholeError)
    def _(msg):
        print(f"ERROR: {msg.message}", file=config.stderr)
        print(msg)

    @output_message.register(Pong)
    def _(msg):
        print(f"Pong: {b16encode(msg.ping_id)}: {msg.time_of_flight}s")

    @output_message.register(CodeAllocated)
    def _(msg):
        print(f"Secret code: {msg.code}")

    @output_message.register(PeerConnected)
    def _(msg):
        nonlocal peer_connected
        peer_connected = True
        nice_verifier = " ".join(
            msg.verifier[a:a+4]
            for a in range(0, len(msg.verifier), 4)
        )
        print(f"Peer is connected.\nVerifier: {nice_verifier}")

    @output_message.register(Listening)
    def _(msg):
        print(f"Listening: {msg.listen}")

    @output_message.register(RemoteListeningSucceeded)
    def _(msg):
        print(f"Peer is listening: {msg.listen} (-> {msg.connect})")

    @output_message.register(RemoteListeningFailed)
    def _(msg):
        print(f"Peer failed to listen: {msg.listen}: {msg.reason}")

    @output_message.register(IncomingConnection)
    def _(msg):
        connections[msg.id] = Connection(msg.endpoint, 0, 0, msg.listener_id)

    @output_message.register(RemoteConnectFailed)
    def _(msg):
        print("Forwarding to {} failed: {}".format(connections[msg.id].endpoint, msg.reason))
        # we will get an OutgoingDone(msg.id) for this one too, so do not delete here!
        # (XXX should maybe mark it as "failed" somehow?)
        ##del connections[msg.id]

    @output_message.register(IncomingLost)
    def _(msg):
        try:
            del connections[msg.id]
            print(f"Lost: {msg.id}: {msg.reason}")
        except KeyError:
            print("WEIRD: got IncomingLost(id={}) but don't have that connection".format(msg.id))

    @output_message.register(IncomingDone)
    def _(msg):
        try:
            del connections[msg.id]
            print(f"{msg.id}: closed")
        except KeyError:
            print("WEIRD: got IncomingDone(id={}) but don't have that connection".format(msg.id))

    @output_message.register(OutgoingLost)
    def _(msg):
        try:
            del connections[msg.id]
            print(f"Lost: {msg.id}: {msg.reason}")
        except KeyError:
            print("WEIRD: got OutgoingLost(id={}) but don't have that connection".format(msg.id))

    @output_message.register(OutgoingDone)
    def _(msg):
        try:
            del connections[msg.id]
            print(f"{msg.id}: closed")
        except KeyError:
            print("WEIRD: got OutgoingDone(id={}) but don't have that connection".format(msg.id))

    @output_message.register(OutgoingConnection)
    def _(msg):
        connections[msg.id] = Connection(msg.endpoint, 0, 0, msg.listener_id)

    @output_message.register(BytesIn)
    def _(msg):
        print(msg)
        connections[msg.id] = Connection(
            connections[msg.id].endpoint,
            connections[msg.id].i + msg.bytes,
            connections[msg.id].o,
            "fixme"
        )

    @output_message.register(BytesOut)
    def _(msg):
        connections[msg.id] = Connection(
            connections[msg.id].endpoint,
            connections[msg.id].i,
            connections[msg.id].o + msg.bytes,
            "fixme"
        )

    @output_message.register(WormholeClosed)
    def _(msg):
        print(f"Closed: {msg.result}")
        we_closed_d.callback(None)

    @output_message.register(Welcome)
    def _(msg):
        print("Connected.")
        if "motd" in msg.welcome:
            print(textwrap.fill(msg.welcome["motd"].strip(), 80, initial_indent="    ", subsequent_indent="    "))

    @output_message.register(GotMessageFromPeer)
    def _(msg):
        peer_msg = json.loads(msg.message)
        print(f"peer: {peer_msg}")
        # XXX part of experimental / PoC "explicit shutdown" flow;
        # probably move to FowlWormhole or similar if useful
        if peer_msg == {"closing": True}:
            fowl_wh._wormhole.send_message(json.dumps({"closed": True}).encode("utf8"))
            we_closed_d.callback(None)
            fowl_wh._wormhole.close()
        elif peer_msg == {"closed": True}:
            print("peer has closed.")
            when_explicitly_closed_d.callback(None)

    fowl_wh = await create_fowl(config, output_message)
    fowl_wh.start()

    ##@daemon.set_trace
    def _(o, i, n):
        print("{} --[ {} ]--> {}".format(o, i, n))

    #XXX kind of experimental / PoC for "close down a session nicely";
    #move to FowlWormhole if actually useful
    async def disconnect_session():
        if peer_connected:
            fowl_wh._wormhole.send_message(json.dumps({"closing": True}).encode("utf8"))
            # (what would {"closed": False} even mean, though?)

            async def wait_for_user():

                def user_got_bored_waiting(*args):
                    # add back original handler, probably Twisted's
                    signal.signal(signal.SIGINT, old_handler)
                    when_explicitly_closed_d.callback(None)
                old_handler = signal.signal(signal.SIGINT, user_got_bored_waiting)

                start = reactor.seconds()

                while (reactor.seconds() - start < 10) and not when_explicitly_closed_d.called:
                    # XXX can we .. catch something here so a second
                    # Ctrl-C means "don't wait, just close"?
                    await deferLater(reactor, 1.1, lambda: None)
                    print('waiting for {"closed": True} from peer')

            await race([
                when_explicitly_closed_d,
                ensureDeferred(wait_for_user()),
            ])

        try:
            await fowl_wh.close_wormhole()
        except wormhole_errors.LonelyError:
            # maybe just say nothing? why does the user care about
            # this? (they probably hit ctrl-c anyway, how else can you get here?)
            print("Wormhole closed without peer")
        print("done")

    ### XXX use PleaseCloseWormhole, approximately
    reactor.addSystemEventTrigger("before", "shutdown", lambda: ensureDeferred(disconnect_session()))

    if config.code is not None:
        fowl_wh.command(
            SetCode(config.code)
        )
    else:
        fowl_wh.command(
            AllocateCode(config.code_length)
        )

    for command in config.commands:
        fowl_wh.command(command)

    last_displayed = None
    while not we_closed_d.called:
        await deferLater(reactor, 1, lambda: None)
        if connections and last_displayed != set(connections.values()):
            for ident in sorted(connections.keys()):
                conn = connections[ident]
                print(f"{ident}: {humanize.naturalsize(conn.i)} in, {humanize.naturalsize(conn.o)} out")
            last_displayed = set(connections.values())


class SubchannelForwarder(Protocol):
    """
    This is the side of the protocol that was listening .. so it has
    opened a subchannel and sent the initial message. So the
    state-machine here is waiting for the reply before forwarding
    data.

    Forwards data to the `.other_protocol` from the Factory.
    """
    m = automat.MethodicalMachine()

    @m.state(initial=True)
    def await_confirmation(self):
        """
        Waiting for the reply from the other side.
        """

    @m.state()
    def evaluating(self):
        """
        Making the local connection.
        """

    @m.state()
    def forwarding_bytes(self):
        """
        All bytes go to the other side.
        """

    @m.state()
    def finished(self):
        """
        We are done (something went wrong or the wormhole session has been
        closed)
        """

    @m.input()
    def got_bytes(self, data):
        """
        received some number of bytes
        """

    @m.input()
    def no_confirmation(self):
        """
        Don't need to wait for a message
        """

    @m.input()
    def got_reply(self, msg):
        """
        Got a complete message (without length prefix)
        """

    @m.input()
    def remote_connected(self):
        """
        The reply message was positive
        """

    @m.input()
    def remote_not_connected(self, reason):
        """
        The reply message was negative
        """

    @m.input()
    def too_much_data(self, reason):
        """
        Too many bytes sent in the reply.
        """
        # XXX is this really an error in this direction? i think only
        # on the other side...

    @m.input()
    def subchannel_closed(self, reason):
        """
        Our subchannel has closed
        """

    @m.output()
    def find_message(self, data):
        """
        Buffer 'data' and determine if a complete message exists; if so,
        inject that input.
        """
        self._buffer += data
        bsize = len(self._buffer)

        if bsize >= 2:
            msgsize, = struct.unpack("!H", self._buffer[:2])
            if bsize > msgsize + 2:
                self.too_much_data()
                return
            elif bsize == msgsize + 2:
                msg = msgpack.unpackb(self._buffer[2:2 + msgsize])
                # this used to have a callLater(0, ..) for the
                # got_reply -- not sure this can happen "in practice",
                # but in testing at least the server can start sending
                # bytes _immediately_ so we must call synchronously..
                return self.got_reply(msg)
        return

    @m.output()
    def check_message(self, msg):
        """
        Initiate the local connection
        """
        if msg.get("connected", False):
            self.remote_connected()
        else:
            self.remote_not_connected(msg.get("reason", "Unknown"))

    @m.output()
    def send_queued_data(self):
        """
        Confirm to the other side that we've connected.
        """
        self.factory.other_proto.transport.resumeProducing()
        self.factory.other_proto._maybe_drain_queue()

    @m.output()
    def emit_remote_failed(self, reason):
        self.factory.message_out(
            RemoteConnectFailed(self.factory.conn_id, reason)
        )

    @m.output()
    def close_connection(self):
        """
        Shut down this subchannel.
        """
        self.transport.loseConnection()

    @m.output()
    def close_other_connection(self):
        """
        Shut down this subchannel.
        """
        self.factory.other_proto.transport.loseConnection()

    @m.output()
    def forward_bytes(self, data):
        """
        Send bytes to the other side
        """
        max_noise = 65000
        while len(data):
            d = data[:max_noise]
            data = data[max_noise:]
            self.factory.message_out(
                BytesOut(self.factory.conn_id, len(d))
            )
            self.factory.other_proto.transport.write(d)

#    @m.output()
#    def emit_incoming_lost(self):
#        # do we need a like "OutgoingLost"?
#        self.factory.message_out(IncomingLost(self.factory.conn_id, "Unknown"))

    await_confirmation.upon(
        no_confirmation,
        enter=forwarding_bytes,
        outputs=[]
    )
    await_confirmation.upon(
        got_bytes,
        enter=await_confirmation,
        outputs=[find_message]
    )
    await_confirmation.upon(
        too_much_data,
        enter=finished,
        outputs=[close_connection]
    )
    await_confirmation.upon(
        got_reply,
        enter=evaluating,
        outputs=[check_message]
    )
    await_confirmation.upon(
        subchannel_closed,
        enter=finished,
        outputs=[],####emit_incoming_lost],
    )

    evaluating.upon(
        remote_connected,
        enter=forwarding_bytes,
        outputs=[send_queued_data]
    )
    evaluating.upon(
        remote_not_connected,
        enter=evaluating,
        outputs=[emit_remote_failed, close_connection]
    )
    evaluating.upon(
        subchannel_closed,
        enter=finished,
        outputs=[]
    )
    forwarding_bytes.upon(
        got_bytes,
        enter=forwarding_bytes,
        outputs=[forward_bytes]
    )
    forwarding_bytes.upon(
        subchannel_closed,
        enter=finished,
        outputs=[close_other_connection]
    )

    finished.upon(
        got_reply,
        enter=finished,
        outputs=[]
    )

    do_trace = m._setTrace

    def connectionMade(self):
        self._buffer = b""
        # self.do_trace(lambda o, i, n: print("{} --[ {} ]--> {}".format(o, i, n)))

        self.local = self.factory.other_proto
        self.factory.other_proto.remote = self
        msg = msgpack.packb({
            "local-destination": self.factory.endpoint_str,
            "listener-id": self.factory.listener_id,
        })
        prefix = struct.pack("!H", len(msg))
        self.transport.write(prefix + msg)

        self.factory.message_out(
            OutgoingConnection(self.factory.conn_id, self.factory.endpoint_str, self.factory.listener_id)
        )

        # MUST wait for reply first -- queueing all data until
        # then
        self.factory.other_proto.transport.pauseProducing()

    def dataReceived(self, data):
        self.got_bytes(data)

    def connectionLost(self, reason):
        if isinstance(reason, ConnectionDone):
            self.factory.message_out(
                OutgoingDone(self.factory.conn_id)
            )
        else:
            self.factory.message_out(
                OutgoingLost(self.factory.conn_id, str(reason))
            )
        self.subchannel_closed(str(reason))
        if self.factory.other_proto:
            self.factory.other_proto.transport.loseConnection()


class ConnectionForward(Protocol):
    """
    The protocol we speak on connections _we_ make to local
    servers. So this basically _just_ forwards bytes.
    """
    m = automat.MethodicalMachine()

    @m.state(initial=True)
    def forwarding_bytes(self):
        """
        All bytes go to the other side.
        """

    @m.state()
    def finished(self):
        """
        We are done (something went wrong or the wormhole session has been
        closed)
        """

    @m.input()
    def got_bytes(self, data):
        """
        received some number of bytes
        """

    @m.input()
    def stream_closed(self, reason):
        """
        The local server has closed our connection
        """

    @m.output()
    def forward_bytes(self, data):
        """
        Send bytes to the other side.

        This will be from the 'actual server' side to our local client
        """
        max_noise = 65000
        while len(data):
            d = data[:max_noise]
            data = data[max_noise:]
            self.factory.message_out(
                BytesIn(self.factory.conn_id, len(d))
            )
            self.factory.other_proto.transport.write(d)

    @m.output()
    def close_other_side(self, reason):
        try:
            if self.factory.other_proto:
                self.factory.other_proto.transport.loseConnection()
        except Exception:
            pass

    @m.output()
    def emit_incoming_done(self, reason):
        if isinstance(reason.value, ConnectionDone):
            self.factory.message_out(
                IncomingDone(self.factory.conn_id)
            )
        else:
            self.factory.message_out(
                IncomingLost(self.factory.conn_id, str(reason))
            )

    forwarding_bytes.upon(
        got_bytes,
        enter=forwarding_bytes,
        outputs=[forward_bytes]
    )
    forwarding_bytes.upon(
        stream_closed,
        enter=finished,
        outputs=[emit_incoming_done, close_other_side]
    )

    def connectionMade(self):
        pass

    def dataReceived(self, data):
        self.got_bytes(data)

    def connectionLost(self, reason):
        self.stream_closed(reason)


class LocalServer(Protocol):
    """
    Listen on an endpoint. On every connection: open a subchannel,
    follow the protocol from _forward_loop above (ultimately
    forwarding data).
    """

    def connectionMade(self):
        self.queue = []
        self.remote = None
        self._conn_id = allocate_connection_id()

        # XXX refactor into SubchannelForwarder.onOpen or whatever, since that exists anyway?
        def got_proto(proto):
            return proto

        # XXX do we need registerProducer somewhere here?
        factory = Factory.forProtocol(SubchannelForwarder)
        factory.other_proto = self
        factory.conn_id = self._conn_id
        factory.listener_id = self.factory.listener_id
        factory.endpoint_str = self.factory.endpoint_str
        factory.config = self.factory.config
        factory.message_out = self.factory.message_out
        # Note: connect_ep here is the Wormhole provided
        # IClientEndpoint that lets us create new subchannels -- not
        # to be confused with the endpoint created from the "local
        # endpoint string"
        d = self.factory.connect_ep.connect(factory)
        d.addCallback(got_proto)

        def err(f):
            self.factory.message_out(
                WormholeError(
                    str(f.value),
                    # extra={"id": self._conn_id}
                )
            )
        d.addErrback(err)
        return d

    def _maybe_drain_queue(self):
        while self.queue:
            msg = self.queue.pop(0)
            self.remote.transport.write(msg)
        self.queue = None

    def connectionLost(self, reason):
        # XXX causes duplice local_close 'errors' in magic-wormhole ... do we not want to do this?)
        if self.remote is not None and self.remote.transport:
            self.remote.transport.loseConnection()

    def dataReceived(self, data):
        self.factory.message_out(
            BytesIn(self._conn_id, len(data))
        )
        self.remote.transport.write(data)


class Incoming(Protocol):
    """
    Handle an incoming Dilation subchannel. This will be from a
    listener on the other end of the wormhole.

    There is an opening message, and then we forward bytes.

    The opening message is a length-prefixed blob; the first 2
    bytes of the stream indicate the length (an unsigned short in
    network byte order).

    The message itself is msgpack-encoded.

    A single reply is produced, following the same format: 2-byte
    length prefix followed by a msgpack-encoded payload.

    The opening message contains a dict like::

        {
            "local-desination": "tcp:localhost:1234",
        }

    The "forwarding" side (i.e the one that opened the subchannel)
    MUST NOT send any data except the opening message until it
    receives a reply from this side. This side (the connecting
    side) may deny the connection for any reason (e.g. it might
    not even try, if policy says not to).

    XXX want some opt-in / permission on this side, probably? (for
    now: anything goes)
    """

    m = automat.MethodicalMachine()
    set_trace = m._setTrace

    @m.state(initial=True)
    def await_message(self):
        """
        The other side must send us a message
        """

    @m.state()
    def local_policy_check(self):
        """
        A connection has come in; we must check our policy
        """

    @m.state()
    def local_connect(self):
        """
        The initial message tells us where to connect locally
        """

    @m.state()
    def forwarding_bytes(self):
        """
        We are connected and have done the proper information exchange;
        now we merely forward bytes.
        """

    @m.state()
    def finished(self):
        """
        Completed the task of forwarding (e.g. client closed connection,
        subchannel closed, fatal error, etc)
        """

    @m.input()
    def policy_bad(self, msg):
        """
        The local policy check has failed (not allowed)
        """

    @m.input()
    def policy_ok(self, msg):
        """
        The local policy check has succeeded (allowed)
        """

    @m.input()
    def got_bytes(self, data):
        """
        We received some bytes
        """

    @m.input()
    def too_much_data(self, reason):
        """
        We received too many bytes (for the first message).
        """

    @m.input()
    def got_initial_message(self, msg):
        """
        The entire initial message is received (`msg` excludes the
        length-prefix but it is not yet parsed)
        """

    @m.input()
    def subchannel_closed(self, reason):
        """
        This subchannel has been closed
        """

    @m.input()
    def connection_made(self):
        """
        We successfully made the local connection
        """

    @m.input()
    def connection_failed(self, reason):
        """
        Making the local connection failed
        """

    @m.output()
    def close_connection(self, reason):
        """
        We wish to close this subchannel
        """
        self.transport.loseConnection()

    @m.output()
    def close_local_connection(self):
        """
        We wish to close this subchannel
        """
        if self._local_connection and self._local_connection.transport:
            self._local_connection.transport.loseConnection()

    @m.output()
    def forward_data(self, data):
        assert self._buffer is None, "Internal error: still buffering"
        assert self._local_connection is not None, "expected local connection by now"
        self._local_connection.transport.write(data)
        self.factory.message_out(
            BytesOut(self._conn_id, len(data))
        )

    @m.output()
    def find_message(self, data):
        """
        Buffer this data and determine if we have a single message yet
        """
        # state-machine shouldn't allow this, but just to be sure
        assert self._buffer is not None, "Internal error: buffer is gone"

        self._buffer += data
        bsize = len(self._buffer)
        if bsize >= 2:
            expected_size, = struct.unpack("!H", self._buffer[:2])
            if bsize >= expected_size + 2:
                first_msg = self._buffer[2:2 + expected_size]
                self._buffer = None
                # there should be no "leftover" data
                if bsize > 2 + expected_size:
                    ##raise RuntimeError("protocol error: more than opening message sent")
                    self.too_much_data("Too many bytes sent")
                    return
                # warning: recursive state-machine message
                self.got_initial_message(msgpack.unpackb(first_msg))

    @m.output()
    def send_negative_reply(self, reason):
        """
        Tell our peer why we're closing them
        """
        self._negative(reason)

    @m.output()
    def send_negative_reply_subchannel(self):
        """
        Tell our peer local policy doesn't let the connection happen.
        """
        pass#self._negative("XXXAgainst local policy")

    def _negative(self, reason):
        msg = msgpack.packb({
            "connected": False,
            "reason": reason,
        })
        prefix = struct.pack("!H", len(msg))
        self.transport.write(prefix + msg)


    @m.output()
    def send_positive_reply(self):
        """
        Reply to the other side that we've connected properly
        """
        # XXX there's other sections like this; factor to pack_netstring() or something
        msg = msgpack.packb({
            "connected": True,
        })
        prefix = struct.pack("!H", len(msg))
        self.transport.write(prefix + msg)

    @m.output()
    def emit_incoming_connection(self, msg):
        self.factory.message_out(
            IncomingConnection(self._conn_id, msg["local-destination"], msg.get("listener-id", None))
        )

    @m.output()
    def do_policy_check(self, msg):
        # note: do this policy check after the IncomingConnection
        # message is emitted (otherwise there will be cases where we
        # emit _just_ a IncomingLost which is confusing)
        if self.factory.config.connect_policy is not None:
            ep = clientFromString(reactor, msg["local-destination"])
            if not self.factory.config.connect_policy.can_connect(ep):
                # warning: synchronous state-machine use
                self.policy_bad(msg)
            else:
                # warning: synchronous state-machine use
                self.policy_ok(msg)

    @m.output()
    def local_disconnect(self):
        self._negative("Against local policy")
        self.transport.loseConnection()

    @m.output()
    def emit_incoming_lost(self, msg):
        self.factory.message_out(
            IncomingLost(
                self._conn_id,
                f"Incoming connection against local policy: {msg['local-destination']}"
            )
        )

    @m.output()
    def establish_local_connection(self, msg):
        """
        FIXME
        """
        ep = clientFromString(reactor, msg["local-destination"])

        factory = Factory.forProtocol(ConnectionForward)
        factory.other_proto = self
        factory.config = self.factory.config
        factory.conn_id = self._conn_id
        factory.message_out = self.factory.message_out

#emit was here

        d = ep.connect(factory)

        def bad(fail):
            # ideally maybe want a @output method send_incoming_lost or so?
            self.factory.message_out(
                IncomingLost(
                    self._conn_id,
                    f"{msg['local-destination']}: {fail.getErrorMessage()}",
                )
            )
            reactor.callLater(0, lambda: self.connection_failed(fail.getErrorMessage()))
            return None

        def assign(proto):
            # if we hit the error-case above, we "handled" it so want
            # to return None -- but this pushes us into the "success"
            # / callback side, albiet with "None" as the value
            self._local_connection = proto
            if self._local_connection is not None:
                reactor.callLater(0, lambda: self.connection_made())
            return proto
        d.addErrback(bad)
        d.addCallback(assign)

    await_message.upon(
        got_bytes,
        enter=await_message,
        outputs=[find_message]
    )
    await_message.upon(
        too_much_data,
        enter=finished,
        outputs=[close_connection]
    )
    await_message.upon(
        got_initial_message,
        enter=local_policy_check,
        outputs=[emit_incoming_connection, do_policy_check]
    )

    local_policy_check.upon(
        policy_ok,
        enter=local_connect,
        outputs=[establish_local_connection]
    )
    local_policy_check.upon(
        policy_bad,
        enter=finished,
        outputs=[local_disconnect, emit_incoming_lost]
    )

    await_message.upon(
        subchannel_closed,
        enter=finished,
        outputs=[],
    )
    local_connect.upon(
        connection_made,
        enter=forwarding_bytes,
        outputs=[send_positive_reply]
    )
    local_connect.upon(
        connection_failed,
        enter=finished,
        outputs=[send_negative_reply, close_connection]
    )
    # this will happen if our policy doesn't allow this port, for example
    local_connect.upon(
        subchannel_closed,
        enter=finished,
        outputs=[],
    )

    forwarding_bytes.upon(
        got_bytes,
        enter=forwarding_bytes,
        outputs=[forward_data]
    )
    forwarding_bytes.upon(
        subchannel_closed,
        enter=finished,
        outputs=[close_local_connection]
    )

    finished.upon(
        subchannel_closed,
        enter=finished,
        outputs=[]  # warning? why does this happen?
    )

    # logically kind-of makes sense, but how are we getting duplicate closes?
    # finished.upon(
    #     subchannel_closed,
    #     enter=finished,
    #     outputs=[],
    # )

    # API methods from Twisted
    # Ideally all these should do is produce some input() to the machine
    def connectionMade(self):
        """
        Twisted API
        """
        self._conn_id = allocate_connection_id()
        # XXX first message should tell us where to connect, locally
        # (want some kind of opt-in on this side, probably)
        self._buffer = b""
        self._local_connection = None
        # def tracer(o, i, n):
        #     print("{} --[ {} ]--> {}".format(o, i, n))
        # self.set_trace(tracer)

    def connectionLost(self, reason):
        """
        Twisted API
        """
#        self.factory.message_out(
#            IncomingLost(self._conn_id, reason)
#        )
        self.subchannel_closed(reason)

    def dataReceived(self, data):
        """
        Twisted API
        """
        self.got_bytes(data)


##XXX needs a shutdown path
## this is currently via "cancel"
## so need a "await FowlDaemon.shutdown()" or similar which can do
## async stuff like shut down all listeners, etc etc

## refactoring _forward_loop etc
##XXX doesn't exist yet but ... @implementer(IWormholeDelegate)


class FowlWormhole:
    """
    Does I/O related to a wormhole connection.
    Co-ordinates between the wormhole, user I/O and the daemon state-machine.
    """

    def __init__(self, reactor, wormhole, daemon, config):
        self._reactor = reactor
        self._listening_ports = []
        self._wormhole = wormhole
        self._config = config
        self._daemon = daemon
        self._done = When() # we have shut down completely
        self._connected = When()  # our Peer has connected
        self._got_welcome = When()  # we received the Welcome from the server
        self.connect_ep = self.control_proto = None

    # XXX wants to be an IService?
    async def stop(self):
        for port in self._listening_ports:
            # note to self port.stopListening and port.loseConnection are THE SAME
            port.stopListening()
        if self.control_proto is not None:
            self.control_proto.transport.loseConnection()
            await self.control_proto.when_done()
        await self._wormhole.close()
        # XXX put "session ending" code here, if it's useful

    # XXX wants to be an IService?
    def start(self):

        # tie "we got a code" into the state-machine
        ensureDeferred(self._wormhole.get_code()).addCallbacks(
            self._daemon.code_allocated,
            self._handle_error,
        )

        # pass on the welcome message
        ensureDeferred(self._wormhole.get_welcome()).addCallbacks(
            lambda hello: self._daemon.got_welcome(hello),
            self._handle_error,
        )

        # when we get the verifier and versions, we emit "peer
        # connected"
        # (note that wormhole will "error" these Deferreds when the
        # wormhole shuts down early, e.g. with LonelyError -- but we
        # handle that elsewhere, so ignore errors here)
        results_d = DeferredList(
            [
                self._wormhole.get_verifier(),
                self._wormhole.get_versions(),
            ],
            consumeErrors=True,
        )

        @results_d.addCallback
        def got_results(results):
            for ok, exc in results:
                if not ok:
                    self._handle_error(exc)
                    return
            verifier = results[0][1]
            versions = results[1][1]
            ###print(f"versions: {versions}")
            # XXX should "do stuff" with this, e.g. per
            # https://github.com/meejah/fowl/issues/45 at least "we
            # speak Fowl Application Protocol" or so

            d = self._do_dilate()
            @d.addCallback
            def did_dilate(arg):
                self._daemon.peer_connected(verifier, versions)
                return arg

        # hook up "incoming message" to input
        def got_message(plaintext):
            self._daemon.got_message(plaintext)
            ensureDeferred(self._wormhole.get_message()).addCallback(
                got_message,
            ).addErrback(self._handle_error)
        ensureDeferred(self._wormhole.get_message()).addCallback(
            got_message,
        ).addErrback(self._handle_error)

        # hook up wormhole closed to input
        def was_closed(why):
            if isinstance(why, str):
                reason = why
            elif isinstance(why.value, wormhole_errors.LonelyError):
                reason = "lonely"
            elif why.value.args:
                reason = why.value.args[0]
            else:
                reason = str(why)
            self._done.trigger(self._reactor, reason)
            self._daemon.shutdown(reason)
        ensureDeferred(self._wormhole._closed_observer.when_fired()).addBoth(was_closed)

    # public API methods

    async def close_wormhole(self):
        """
        Shut down the wormhole
        """
        # XXX see also the whole "how to shutdown half-close etc"
        await self._wormhole.close()
        # once the wormhole "actually" closes, the state-machine will
        # trigger our "stop" codepath

    def command(self, command: FowlCommandMessage) -> None:
        """
        Process an incoming command
        """

        @functools.singledispatch
        async def cmd(msg):
            print(f"Unprocessed command: {msg}")

        @cmd.register(AllocateCode)
        async def _(msg):
            if msg.length:
                self._wormhole.allocate_code(msg.length)
            else:
                self._wormhole.allocate_code()

        @cmd.register(SetCode)
        async def _(msg):
            self._wormhole.set_code(msg.code)

        @cmd.register(LocalListener)
        async def _(msg):
            await self.when_connected()
            assert self.connect_ep is not None, "need connect ep"
            await _local_to_remote_forward(self._reactor, self._config, self.connect_ep, self._listening_ports.append, self._daemon._message_out, msg)
            # XXX cheating? private access (_daemon._message_out)

        @cmd.register(RemoteListener)
        async def _(msg):
            await self.when_connected()
            assert self.control_proto is not None, "need control_proto"
            msg = await _remote_to_local_forward(self.control_proto, msg)
            self._daemon._message_out(msg)
            # XXX cheating? private access (_daemon._message_out)

        @cmd.register(GrantPermission)
        async def _(msg):
            self._config.connect_policy.ports.extend(msg.connect)
            self._config.listen_policy.ports.extend(msg.listen)
            print(
                "Permission granted. Listen={}, Connect={}".format(
                    self._config.listen_policy,
                    self._config.connect_policy,
                ),
                file=self._config.stderr,
                flush=True,
            )

        @cmd.register(DangerDisablePermissionCheck)
        async def _(msg):
            try:
                self._config.connect_policy = AnyConnectPolicy()
                self._config.listen_policy = AnyListenPolicy()
                print("DANGER. All connect / listen endpoints allowed.", file=self._config.stderr, flush=True)
            except Exception as e:
                print("BAD", type(e))

        @cmd.register(Ping)
        async def _(msg):
            if hasattr(self._wormhole, "_boss"):
                if hasattr(self._wormhole._boss, "_D"):
                    if self._wormhole._boss._D._manager is not None:
                        def got_pong(round_trip):
                            self._daemon._message_out(Pong(msg.ping_id, round_trip))
                        self._wormhole._boss._D._manager.send_ping(msg.ping_id, got_pong)
                    else:
                        raise Exception("Cannot send ping: not in Dilation")
                else:
                    raise Exception("Cannot send ping: no Dilation manager")
            else:
                raise Exception("Cannot send ping: no boss")

        ensureDeferred(cmd(command)).addErrback(self._handle_error)

    # our own callbacks / notifications

    async def get_welcome(self):
        return await self._got_welcome.when_triggered()

    def when_done(self):
        return self._done.when_triggered()

    def when_connected(self):
        """
        Our peer has connected.

        Fires with the "verifier" of the peer (this is an optional
        measure whereby users may compare these; if they're identical
        they are communicating with the intended partner)
        """
        return self._connected.when_triggered()

    def _do_dilate(self):
        self.control_ep, self.connect_ep, self.listen_ep = self._wormhole.dilate(
            transit_relay_location="tcp:magic-wormhole-transit.debian.net:4001",
        )
        d = ensureDeferred(self._post_dilation_setup())
        d.addErrback(self._handle_error)
        return d

    def _handle_error(self, f):
        # hmm, basically any of the "wormhole callbacks" we asked for
        # stuff will definitely "errback" when we end the wormhole --
        # sometimes with WormholeClosed, or LonelyError -- but also we
        # should feedback those into the state-machine ONLY from one,
        # like get_message (?) -- but what do we do with the others?
        if isinstance(f.value, (wormhole_errors.WormholeClosed, wormhole_errors.LonelyError)):
            pass
        else:
            self._report_error(f.value)

    def _report_error(self, e):
        self._daemon._message_out(
            WormholeError(str(e))
        )

    async def _post_dilation_setup(self):
        # listen for commands from the other side on the control channel
        assert self.control_ep is not None, "Need control connection"
        fac = Factory.forProtocol(Commands)
        fac.config = self._config
        fac.message_out = self._daemon._message_out  # XXX
        fac.reactor = self._reactor
        fac.connect_ep = self.connect_ep  # so we can pass it command handlers
        self.control_proto = await self.control_ep.connect(fac)

        # listen for incoming subchannel OPENs
        in_factory = Factory.forProtocol(Incoming)
        in_factory.config = self._config
        in_factory.connect_ep = self.connect_ep
        in_factory.message_out = self._daemon._message_out
        listen_port = await self.listen_ep.listen(in_factory)
        self._listening_ports.append(listen_port)

        await self._wormhole.get_unverified_key()
        verifier_bytes = await self._wormhole.get_verifier()  # might WrongPasswordError
        self._connected.trigger(self._reactor, verifier_bytes)



# FowlDaemon is the state-machine
#  - ultimately, it goes some notifications from 'the wormhole' but only via the I/O thing
#  - no "async def" / Deferred-returning methods
#  - no I/O
#
# FowlWormhole is "the I/O thing"
#  - can do async
#  - can do I/O
#  - proxies I/O and async'ly things

from .daemon import FowlDaemon  # noqa: E402


def maybe_int(i):
    if i is None:
        return None
    return int(i)


# XXX FIXME
# NOTES
# 
# since we want "fowl" to use "fowld" we have two options:
# - start a subprocess and use stdin/out
# - sans-io style (send "messages" in / out of _forward_loop or so)
#
# Would like the second; so we can interact in unit-tests (or here)
# via parsed commands. e.g. we have an AGT union-type, and every
# input-message is a class
def fowld_command_to_json(msg: FowlCommandMessage) -> dict:
    """
    Turn the given `msg` into a corresponding JSON-friendly dict
    (suitable for json.dumps() for example)
    """

    js = asdict(msg)
    # XXX maybe a @singledispatch would give better feedback when we
    # miss one...
    @functools.singledispatch
    def output_command(msg):
        raise RuntimeError(f"Unhandled message: {msg}")

    @output_command.register(GrantPermission)
    def _(msg):
        js["kind"] = "grant-permission"

    @output_command.register(DangerDisablePermissionCheck)
    def _(msg):
        js["kind"] = "danger-disable-permission-check"

    @output_command.register(LocalListener)
    def _(msg):
        js["kind"] = "local"
        js["listen"] = msg.listen
        js["connect"] = msg.connect

    @output_command.register(RemoteListener)
    def _(msg):
        js["kind"] = "remote"
        js["listen"] = msg.listen
        js["connect"] = msg.connect

    @output_command.register(AllocateCode)
    def _(msg):
        js["kind"] = "allocate-code"
        js["length"] = getattr(msg, "length", None)

    @output_command.register(SetCode)
    def _(msg):
        js["kind"] = "set-code"
        js["code"] = msg.code

    @output_command.register(Ping)
    def _(msg):
        js["kind"] = "ping"
        js["ping_id"] = msg.ping_id

    output_command(msg)
    return js


def parse_fowld_command(json_str: str) -> FowlCommandMessage:
    """
    Parse the given JSON message assuming is a command for fowld
    """
    cmd = json.loads(json_str)
    try:
        kind = cmd["kind"]
        del cmd["kind"]
    except KeyError:
        raise ValueError("No 'kind' in command")

    def parser(cls, item_parsers, optional_parsers):
        def parse(js):
            args = {}
            for k, process in item_parsers:
                try:
                    args[k] = js[k] if process is None else process(js[k])
                except KeyError:
                    raise ValueError('"{}" is missing'.format(k))
            for k, process in optional_parsers:
                try:
                    args[k] = js[k] if process is None else process(js[k])
                except KeyError:
                    pass
            return cls(**args)
        return parse

    def is_valid_port(port):
        if isinstance(port, int) and port >= 1 and port < 65536:
            return port
        raise ValueError(f"Invalid port: {port}")

    def port_list(proposed):
        if isinstance(proposed, list):
            return [
                is_valid_port(port) for port in proposed
            ]
        raise ValueError("Port-list must be a list of ints")

    kind_to_message = {
        "allocate-code": parser(AllocateCode, [], [("length", maybe_int)]),
        "set-code": parser(SetCode, [("code", None)], []),
        "local": parser(LocalListener, [("listen", None), ("connect", None)], []),
        "remote": parser(RemoteListener, [("listen", None), ("connect", None)], []),
        "grant-permission": parser(GrantPermission, [("listen", port_list), ("connect", port_list)], []),
        "danger-disable-permission-check": parser(DangerDisablePermissionCheck, [], []),
        "ping": parser(Ping, [("ping_id", None)], []),
    }
    return kind_to_message[kind](cmd)


def fowld_output_to_json(msg: FowlOutputMessage) -> dict:
    """
    Turn the given `msg` into a corresponding JSON-friendly dict
    (suitable for json.dumps() for example)
    """
    # XXX maybe a @singledispatch would give better feedback when we
    # miss one...
    js = asdict(msg)
    js["kind"] = {
        Welcome: "welcome",
        WormholeClosed: "closed",
        CodeAllocated: "code-allocated",
        PeerConnected: "peer-connected",
        Listening: "listening",
        RemoteListeningFailed: "remote-listening-failed",
        RemoteListeningSucceeded: "remote-listening-succeeded",
        RemoteConnectFailed: "remote-connect-failed",
        OutgoingConnection: "outgoing-connection",
        OutgoingLost: "outgoing-lost",
        OutgoingDone: "outgoing-done",
        IncomingConnection: "incoming-connection",
        IncomingLost: "incoming-lost",
        IncomingDone: "incoming-done",
        BytesIn: "bytes-in",
        BytesOut: "bytes-out",
        WormholeError: "error",
        Pong: "pong",
    }[type(msg)]
    return js


def parse_fowld_output(json_str: str) -> FowlOutputMessage:
    """
    Parse the given JSON message assuming it came from fowld.
    :raises: ValueError if it's an invalid message.
    """
    cmd = json.loads(json_str)
    try:
        kind = cmd["kind"]
        del cmd["kind"]
    except KeyError:
        raise ValueError("No 'kind' in command")

    def parser(cls, item_parsers):
        def parse(js):
            args = {}
            for k, process in item_parsers:
                try:
                    args[k] = js[k] if process is None else process(js[k])
                except KeyError:
                    raise ValueError(f'"{k}" is missing')
            return cls(**args)
        return parse

    kind_to_message = {
        "welcome": parser(Welcome, [("url", None), ("welcome", None)]),
        "wormhole-closed": parser(WormholeClosed, [("result", None)]),
        "allocate-code": parser(AllocateCode, [("length", maybe_int)]),
        "set-code": parser(SetCode, [("code", None)]),
        "code-allocated": parser(CodeAllocated, [("code", None)]),
        "peer-connected": parser(PeerConnected, [("verifier", binascii.unhexlify), ("versions", None)]),
        "listening": parser(Listening, [("listener_id", None), ("listen", None), ("connect", None)]),
        "remote-listening-failed": parser(RemoteListeningFailed, [("listen", None), ("reason", None)]),
        "remote-listening-succeeded": parser(RemoteListeningSucceeded, [("listen", None), ("connect", None), ("listener_id", None)]),
        "remote-connect-failed": parser(RemoteConnectFailed, [("id", int), ("reason", None)]),
        "outgoing-connection": parser(OutgoingConnection, [("id", int), ("endpoint", None), ("listener_id", None)]),
##        "outgoing-lost": parser(),
        "outgoing-done": parser(OutgoingDone, [("id", int)]),
        "incoming-connection": parser(IncomingConnection, [("id", int), ("endpoint", None), ("listener_id", None)]),
        "incoming-lost": parser(IncomingLost, [("id", int), ("reason", None)]),
        "incoming-done": parser(IncomingDone, [("id", int)]),
        "bytes-in": parser(BytesIn, [("id", int), ("bytes", int)]),
        "bytes-out": parser(BytesOut, [("id", int), ("bytes", int)]),
        "closed": parser(WormholeClosed, [("result", str)]),
        "pong": parser(Pong, [("ping_id", bytes), ("time_of_flight", float)]),
    }
    return kind_to_message[kind](cmd)


async def create_fowl(config, output_fowl_message):
    w = await wormhole_from_config(reactor, config)

    if config.debug_file:
        kind = "invite" if config.code is None else "accept"
        w.debug_set_trace(kind, which="B N M S O K SK R RC L C T", file=config.debug_file)

    def command_message(msg):
        if isinstance(msg, PleaseCloseWormhole):
            d = ensureDeferred(fowl.close_wormhole())
            d.addErrback(lambda f: print(f"Error closing: {f.value}"))

    sm = FowlDaemon(config, output_fowl_message, command_message)

#    @sm.set_trace
    def _(start, edge, end):
        print(f"trace: {start} --[ {edge} ]--> {end}")
    fowl = FowlWormhole(reactor, w, sm, config)
    return fowl


async def forward(reactor, config):
    """
    Set up a wormhole and process commands relating to forwarding.

    That is, run the main loop of the forward:
       - perform setup (version stuff etc)
       - wait for commands (as single-line JSON messages) on stdin
       - write results to stdout (as single-line JSON messages)
       - service subchannels

    See docs/messages.rst for more
    """
    def output_fowl_message(msg):
        js = fowld_output_to_json(msg)
        print(
            json.dumps(js),
            file=config.stdout,
            flush=True,
        )

    fowl = await create_fowl(config, output_fowl_message)
    fowl.start()

    # arrange to read incoming commands from stdin
    create_stdio = config.create_stdio or StandardIO
    dispatch = LocalCommandDispatch(config, fowl)
    create_stdio(dispatch)

    try:
        await Deferred()
    except CancelledError:
        await fowl.stop()


async def _local_to_remote_forward(reactor, config, connect_ep, on_listen, on_message, cmd):
    """
    Listen locally, and for each local connection create an Outgoing
    subchannel which will connect on the other end.
    """
    # XXX these lines are "uncovered" but we clearly run them ... so
    # something wrong with subprocess coverage?? again???
    ep = serverFromString(reactor, cmd.listen)

    if config.listen_policy is not None:
        if not config.listen_policy.can_listen(ep):
            raise ValueError('Policy doesn\'t allow listening on "{}"'.format(ep._port))

    listener_id = b16encode(os.urandom(3)).decode("ascii")
    factory = Factory.forProtocol(LocalServer)
    factory.config = config
    factory.listener_id = listener_id
    factory.endpoint_str = cmd.connect
    factory.connect_ep = connect_ep
    factory.message_out = on_message
    port = await ep.listen(factory)
    on_listen(port)
    on_message(Listening(listener_id, cmd.listen, cmd.connect))


async def _remote_to_local_forward(control_proto, cmd):
    """
    Ask the remote side to listen on a port, forwarding back here
    (where our Incoming subchannel will be told what to connect
    to).
    """
    reply = await control_proto.request_forward(cmd)
    return reply


class Commands(Protocol):
    """
    Listen for (and send) commands over the command subchannel
    """

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._ports = []
        self._done = When()
        # outstanding requests, awaiting replies
        self._remote_requests = dict()
        self._local_requests = dict()
        self._create_request_id = count(1)

    def request_forward(self, cmd):
        rid = next(self._create_request_id)
        msg = msgpack.packb({
            "kind": "remote-to-local",
            "listen-endpoint": cmd.listen,
            "connect-endpoint": cmd.connect,
            "reply-id": rid,
        })
        d = Deferred()
        self._local_requests[rid] = d, cmd
        prefix = struct.pack("!H", len(msg))
        self.transport.write(prefix + msg)
        return d

    def when_done(self):
        return self._done.when_triggered()

    def dataReceived(self, data):
        # XXX can we depend on data being "one message"? or do we need length-prefixed?
        bsize = len(data)
        assert bsize >= 2, "expected at least 2 bytes"
        expected_size, = struct.unpack("!H", data[:2])
        assert bsize == expected_size + 2, "data has more than the message: {} vs {}: {}".format(bsize, expected_size + 2, repr(data[:55]))
        msg = msgpack.unpackb(data[2:])
        if msg["kind"] == "listener-response":
            rid = int(msg["reply-id"])
            d, original_cmd = self._local_requests[rid]
            del self._local_requests[rid]

            if msg["listening"]:
                d.callback(RemoteListeningSucceeded(msg.get("listener-id"), original_cmd.listen, original_cmd.connect))
            else:
                d.callback(RemoteListeningFailed(original_cmd.listen, msg.get("reason", "Missing reason")))

        elif msg["kind"] == "remote-to-local":
            rid = int(msg["reply-id"])
            assert rid >= 1 and rid < 2**53, "reply-id out of range"
            self._remote_requests[rid] = None

            listen_ep = serverFromString(reactor, msg["listen-endpoint"])

            if self.factory.config.listen_policy is not None:
                if not self.factory.config.listen_policy.can_listen(listen_ep):
                    self._reply_negative(rid, "Against local policy")
                    self.transport.loseConnection()
                    return

            #XXX FIXME why did i have to add 'listener_id' code in two places??
            listener_id = b16encode(os.urandom(3)).decode("ascii")
            factory = Factory.forProtocol(LocalServer)
            factory.config = self.factory.config
            factory.message_out = self.factory.message_out
            factory.connect_ep = self.factory.connect_ep
            factory.endpoint_str = msg["connect-endpoint"]
            factory.listener_id = listener_id

            # XXX this can fail (synchronously) if address in use (e.g.)
            d = listen_ep.listen(factory)
            self._remote_requests[rid] = d

            def got_port(port):
                self._reply_positive(rid, listener_id)
                self.factory.message_out(
                    Listening(
                        listener_id,
                        msg["listen-endpoint"],
                        msg["connect-endpoint"],
                    )
                )
                self._ports.append(port)
                return port

            def error(f):
                self._reply_negative(rid, f.getErrorMessage())
                self.factory.message_out(
                    WormholeError(
                        'Failed to listen on "{}": {}'.format(
                            msg["listen-endpoint"],
                            f.value,
                        )
                    )
                )
            d.addCallback(got_port)
            d.addErrback(error)
            # XXX should await port.stopListening() somewhere...at the appropriate time
        else:
            self.factory.message_out(
                WormholeError(
                    "Unknown control command: {msg[kind]}",
                )
            )

    def _reply_positive(self, rid, listener_id):
        """
        Send a positive reply to a remote request
        """
        return self._reply_generic(rid, listening=True, listener_id=listener_id)

    def _reply_negative(self, rid, reason=None):
        """
        Send a negative reply to a remote request
        """
        return self._reply_generic(rid, listening=False, reason=reason)

    def _reply_generic(self, rid, listening, reason=None, listener_id=None):
        """
        Send a positive or negative reply to a remote request
        """
        if rid in self._remote_requests:
            content = {
                "kind": "listener-response",
                "reply-id": rid,
                "listener-id": listener_id,
                "listening": bool(listening),
            }
            if reason is not None and not listening:
                content["reason"] = str(reason)
            reply = msgpack.packb(content)
            prefix = struct.pack("!H", len(reply))
            self.transport.write(prefix + reply)
            del self._remote_requests[rid]
        else:
            raise ValueError(
                "Replied to '{}' but it is not outstanding".format(rid)
            )

    async def _unregister_ports(self):
        unreg = self._ports
        self._ports = []
        for port in unreg:
            # "might return Deferred" sucks...
            d = port.stopListening()
            if d is not None:
                await d

    def connectionLost(self, reason):
        d = ensureDeferred(self._unregister_ports())

        @d.addCallback
        def _(_):
            self._done.trigger(self.factory.reactor, None)


class LocalCommandDispatch(LineReceiver):
    """
    Wait for incoming commands (as lines of JSON) and dispatch them.
    """
    delimiter = b"\n"

    def __init__(self, cfg, fowl):
        super(LocalCommandDispatch, self).__init__()
        self.config = cfg
        self.fowl = fowl

    def connectionMade(self):
        pass

    def lineReceived(self, line):
        try:
            cmd = parse_fowld_command(line)
            self.fowl.command(cmd)
        except Exception as e:
            print(f"{line.strip()}: failed: {e}")


async def get_tor(
        reactor,
        tor_control_port=None,
        stderr=sys.stderr):
    """
    Create an ITorManager suitable for use with wormhole.create()

    This will attempt to connect to well-known ports and failing that
    will launch its own tor subprocess.
    """
    from wormhole._interfaces import ITorManager

    try:
        import txtorcon
    except ImportError:
        raise click.UsageError(
            'Cannot find txtorcon library (try "pip install txtorcon")'
        )

    # Connect to an existing Tor, or create a new one. If we need to
    # launch an onion service, then we need a working control port (and
    # authentication cookie). If we're only acting as a client, we don't
    # need the control port.

    # we could add a way for the user to configure the Tor endpoint if
    # desired .. but for now we just try the usual suspects, then
    # launch one if that fails.

    # control_ep = clientFromString(reactor, tor_control_port)
    # with all defaults, tries:
    # unix:/var/run/tor/control, localhost:9051, localhost:9151
    try:
        tor = await txtorcon.connect(reactor)
    except Exception as e:
        print(
            f"Failed to connect to Tor: {e}\nAttempting to run Tor",
            file=stderr,
            flush=True,
        )

        def progress(done, tag, summary):
            print(f"Tor: {done}: {tag}: {summary}", file=stderr)

        try:
            tor = await txtorcon.launch(
                reactor,
                progress_updates=progress,
                # data_directory=,
                # tor_binary=,
            )
        except Exception as e:
            raise click.UsageError(
                f"Failed to connect to Tor, then failed to launch: {e}",
            )

    directlyProvides(tor, ITorManager)
    return tor



import os
import random
import string
import json
from io import StringIO

from attr import define

import pytest
import pytest_twisted

from twisted.internet.task import deferLater
from twisted.internet.defer import ensureDeferred, CancelledError
from twisted.internet.protocol import Protocol
from twisted.internet.endpoints import serverFromString, clientFromString

from fowl._proto import _Config, forward
from fowl.observer import Accumulate
from fowl.test.util import ServerFactory, ClientFactory
from fowl.policy import AnyConnectPolicy, AnyListenPolicy


# XXX ultimately we might want a "TestingWormhole" object or something
# to put into wormhole proper.
# It would be in-memory and hook up all protocols to .. itself?


def create_wormhole_factory():

    def stream_of_valid_codes():
        for number in range(1, 1000):
            code = "{}-{}-{}".format(
                number,
                random.choice(string.ascii_letters),
                random.choice(string.ascii_letters),
            )
            yield code

    wormholes = []
    codes = stream_of_valid_codes()

    async def memory_wormhole(cfg):
        print("memory wormhole", cfg)

        @define
        class Endpoint:
            connects: list = []
            listens: list = []

            async def connect(self, addr):
                print("connect", addr)
                return self.connects.pop(0)

            def listen(self, factory):
                print("listen", factory)
                ear = self.listens.pop(0)
                return ear(factory)

        @define
        class Wormhole:
            code: str = None
            control_ep: Endpoint = Endpoint()
            connect_ep: Endpoint = Endpoint()
            listen_ep: Endpoint = Endpoint()

            async def get_welcome(self):
                return {
                    "testing": "this is a testing wormhole",
                }

            def allocate_code(self, words):
                self.code = next(codes)
                return self.code

            async def get_code(self):
                return self.code

            async def get_unverified_key(self):
                return b"0" * 32

            async def get_verifier(self):
                return b"x" * 32

            def dilate(self):
                return (self.control_ep, self.connect_ep, self.listen_ep)

        w = Wormhole()
        wormholes.append(w)
        return w
    return memory_wormhole


async def sleep(reactor, t):
    await deferLater(reactor, t, lambda: None)


@pytest_twisted.ensureDeferred
async def find_message(reactor, config, kind=None, timeout=10, show_error=True):
    """
    Await a message of particular kind in the stdout of config

    :param bool show_error: when True, immediately print any
        "kind=error" messages encountered
    """
    for _ in range(timeout):
        messages = [
            json.loads(line)
            for line in config.stdout.getvalue().split("\n")
            if line
        ]
        for msg in messages:
            if msg["kind"] == kind:
                return msg
            if msg["kind"] == "error":
                print(f"error: {msg.message}")
        await sleep(reactor, 1)
        if False:
            print("no '{}' yet: {}".format(kind, " ".join([m.get("kind", "<kind missing>") for m in messages])))
    raise RuntimeError(
        f"Waited {timeout}s for message of kind={kind}"
    )


class FakeStandardIO(object):
    def __init__(self, proto, reactor, messages):
        self.disconnecting = False  ## XXX why? this is in normal one?
        self.proto = proto
        self.reactor = reactor
        self.messages = messages
        self.proto.makeConnection(self)
        for msg in messages:
            assert isinstance(msg, bytes), "messages must be bytes"
            self.proto.dataReceived(msg)


def ignore_cancel(f):
    if f.trap(CancelledError):
        return None
    return f


# maybe Hypothesis better, via strategies.binary() ?
@pytest_twisted.ensureDeferred
@pytest.mark.parametrize("datasize", [2**6])#range(2**6, 2**17, 2**15))
@pytest.mark.parametrize("who", [True, False])
async def test_forward(reactor, request, mailbox, datasize, who):

    stdios = [
        None,
        None,
    ]

    def create_stdin0(proto, reactor=None):
        stdios[0] = FakeStandardIO(proto, reactor, messages=[])
        return stdios[0]

    def create_stdin1(proto, reactor=None):
        stdios[1] = FakeStandardIO(proto, reactor, messages=[])
        return stdios[1]

    config0 = _Config(
        relay_url=mailbox.url,
        use_tor=False,
        create_stdio=create_stdin0,
        stdout=StringIO(),
        listen_policy=AnyListenPolicy(),
        connect_policy=AnyConnectPolicy(),
    )
    # note: would like to get rid of this ensureDeferred, but it
    # doesn't start "running" the coro until we do this...
    d0 = ensureDeferred(forward(reactor, config0))
    d0.addErrback(ignore_cancel)
    msg = await find_message(reactor, config0, kind="welcome")
    stdios[0].proto.dataReceived(
        json.dumps({
            "kind": "allocate-code",
        }).encode("utf8") + b"\n"
    )

    msg = await find_message(reactor, config0, kind="code-allocated")
    assert 'code' in msg, "Missing code"

    config1 = _Config(
        relay_url=mailbox.url,
        use_tor=False,
        create_stdio=create_stdin1,
        stdout=StringIO(),
        code=msg["code"],
        listen_policy=AnyListenPolicy(),
        connect_policy=AnyConnectPolicy(),
    )

    d1 = ensureDeferred(forward(reactor, config1))
    d1.addErrback(ignore_cancel)
    msg = await find_message(reactor, config1, kind="welcome")
    stdios[1].proto.dataReceived(
        json.dumps({
            "kind": "set-code",
            "code": config1.code,
        }).encode("utf8") + b"\n"
    )

    await find_message(reactor, config0, kind="peer-connected")
    await find_message(reactor, config1, kind="peer-connected")

    class Server(Protocol):
        _message = Accumulate(b"")

        def dataReceived(self, data):
            self._message.some_results(reactor, data)

        async def next_message(self, expected_size):
            return await self._message.next_item(reactor, expected_size)

        def send(self, data):
            self.transport.write(data)


    class Client(Protocol):
        _message = Accumulate(b"")

        def dataReceived(self, data):
            self._message.some_results(reactor, data)

        async def next_message(self, expected_size):
            return await self._message.next_item(reactor, expected_size)

        def send(self, data):
            self.transport.write(data)

    listener = ServerFactory(reactor)
    server_port = await serverFromString(reactor, "tcp:1111").listen(listener)

    # both sides are connected -- now we can issue a "remote listen"
    # request
    stdios[0].proto.dataReceived(
        json.dumps({
            "kind": "local",
            "listen": "tcp:7777",
            "connect": "tcp:localhost:1111",
        }).encode("utf8") + b"\n"
    )
    msg = await find_message(reactor, config0, kind="listening")

    # if we do 'too many' test-cases debian complains about
    # "twisted.internet.error.ConnectBindError: Couldn't bind: 24: Too
    # many open files."
    # gc.collect() doesn't fix it.
    client = clientFromString(reactor, "tcp:localhost:7777") # NB: same port as in "kind=local" message!
    client_proto = await client.connect(ClientFactory(reactor))
    server = await listener.next_client()

    def cleanup():
        d0.cancel()
        d1.cancel()
        client_proto.transport.loseConnection()
        server.transport.loseConnection()
        server_port.stopListening()
    request.addfinalizer(cleanup)

    data = os.urandom(datasize)
    if who:
        client_proto.send(data)
        msg = await server.next_message(len(data))
    else:
        server.send(data)
        msg = await client_proto.next_message(len(data))
    who = not who
    assert msg == data, "Incorrect data transfer"


@pytest_twisted.ensureDeferred
@pytest.mark.parametrize("datasize", [2**6])##range(2**6, 2**17, 2**15))
@pytest.mark.parametrize("who", [True, False])
@pytest.mark.parametrize("wait_peer", [True, False])
async def test_drawrof(reactor, request, mailbox, datasize, who, wait_peer):

    stdios = [
        None,
        None,
    ]

    def create_stdin0(proto, reactor=None):
        stdios[0] = FakeStandardIO(proto, reactor, messages=[])
        return stdios[0]

    def create_stdin1(proto, reactor=None):
        stdios[1] = FakeStandardIO(proto, reactor, messages=[])
        return stdios[1]

    config0 = _Config(
        relay_url=mailbox.url,
        use_tor=False,
        create_stdio=create_stdin0,
        stdout=StringIO(),
        listen_policy=AnyListenPolicy(),
        connect_policy=AnyConnectPolicy(),
    )
    # note: would like to get rid of this ensureDeferred, but it
    # doesn't start "running" the coro until we do this...
    d0 = ensureDeferred(forward(reactor, config0))
    d0.addErrback(ignore_cancel)
    msg = await find_message(reactor, config0, kind="welcome")

    # when connected, issue a "open listener" to one side
    stdios[0].proto.dataReceived(
        json.dumps({
            "kind": "allocate-code",
        }).encode("utf8") + b"\n"
    )

    msg = await find_message(reactor, config0, kind="code-allocated")
    assert 'code' in msg, "Missing code"

    config1 = _Config(
        relay_url=mailbox.url,
        use_tor=False,
        create_stdio=create_stdin1,
        stdout=StringIO(),
        code=msg["code"],
        listen_policy=AnyListenPolicy(),
        connect_policy=AnyConnectPolicy(),
    )
    d1 = ensureDeferred(forward(reactor, config1))
    d1.addErrback(ignore_cancel)
    msg = await find_message(reactor, config1, kind="welcome")

    # now we can set the code on this side
    stdios[1].proto.dataReceived(
        json.dumps({
            "kind": "set-code",
            "code": config1.code,
        }).encode("utf8") + b"\n"
    )

    listener = ServerFactory(reactor)
    server_port = await serverFromString(reactor, "tcp:3333:interface=localhost").listen(listener)
    request.addfinalizer(server_port.stopListening)

    # whether we explicitly wait for our peer, the underlying fowl
    # code should "do the right thing" if we just start issuing
    # listen/etc commands
    if wait_peer:
        print("Explicitly awaiting peers")
        msg = await find_message(reactor, config0, kind="peer-connected")
        msg = await find_message(reactor, config1, kind="peer-connected")
        print("Both sides have a peer")
    else:
        print("Not waiting for peer")

    # both sides are connected -- now we can issue a "remote listen"
    # request
    stdios[0].proto.dataReceived(
        json.dumps({
            "kind": "remote",
            "listen": "tcp:8888",
            "connect": "tcp:localhost:3333",
        }).encode("utf8") + b"\n"
    )

    msg = await find_message(reactor, config1, kind="listening")
    print(f"Listening: {msg}")

    # if we do 'too many' test-cases debian complains about
    # "twisted.internet.error.ConnectBindError: Couldn't bind: 24: Too
    # many open files."
    # gc.collect() doesn't fix it.
    client = clientFromString(reactor, "tcp:localhost:8888")  # NB: same as remote-endpoint
    client_proto = await client.connect(ClientFactory(reactor))
    server = await listener.next_client()

    def cleanup():
        d0.cancel()
        d1.cancel()
        server.transport.loseConnection()
        pytest_twisted.blockon(ensureDeferred(server.when_closed()))
    request.addfinalizer(cleanup)

    data = os.urandom(datasize)
    if who:
        print("client sending")
        client_proto.send(data)
        msg = await server.next_message(len(data))
    else:
        print("server sending")
        server.send(data)
        msg = await client_proto.next_message(len(data))
    assert msg == data, "Incorrect data transfer"



import os

import click


@click.group()
def wizard_gardens():
    """
    Experiments towards a unified Magic Wormhole UI and glue-code
    runner for local services across the network
    """
    pass


@wizard_gardens.command()
def polkit():
    """
    Can we ask for authentication via dbus?
    """
    import dbus
    #session = dbus.SessionBus()
    session = dbus.SystemBus()
    pk = session.get_object("org.freedesktop.PolicyKit1", "/org/freedesktop/PolicyKit1/Authority")
    print(pk)
    subject = [
        "unix-process",
         dict(
             pid=os.getpid(),
         ),
    ]
    action = "org.freedesktop.Flatpak.app-install"
    details = {}
    flags = {}
    cancellation_id = "cancel-me"
    foo = dbus.Interface(pk, dbus_interface="org.freedesktop.PolicyKit1.Authority")
    out = foo.CheckAuthorization(subject, action, details, flags, cancellation_id)
    print(out)
    if False:
        out = foo.EnumerateActions("org.freedesktop.PolicyKit1.Authority")
        names = sorted(i[0] for i in out)
        for x in names:
            print(x)




import click
import pkg_resources

from twisted.internet.task import react
from twisted.internet.defer import ensureDeferred

from wormhole.cli.public_relay import (
    RENDEZVOUS_RELAY as PUBLIC_MAILBOX_URL,
)

from ._proto import (
    _Config,
    forward,
    frontend_accept_or_invite,
    WELL_KNOWN_MAILBOXES,
)
from ._tui import frontend_tui
from .messages import (
    LocalListener,
    RemoteListener,
)
from .policy import (
    LocalhostTcpPortsListenPolicy,
    LocalhostTcpPortsConnectPolicy,
    ArbitraryAddressTcpConnectPolicy,
    ArbitraryInterfaceTcpPortsListenPolicy,
    is_localhost,
)


@click.option(
    "--ip-privacy/--clearnet",
    default=False,
    help="Enable operation over Tor (default is public Internet)",
)
@click.option(
    "--mailbox",
    default=PUBLIC_MAILBOX_URL,
    help='URL for the mailbox server to use (or "default", "local" or "winden" to use well-known servers)',
    metavar="URL or NAME",
)
@click.option(
    "--debug",
    default=None,
    help="Output wormhole state-machine transitions to the given file",
    type=click.File("w", encoding="utf8"),
)
@click.command()
@click.pass_context
def fowld(ctx, ip_privacy, mailbox, debug):
    """
    Forward Over Wormhole, Locally, Daemon

    Low-level daemon to set up and forward streams over Dilated magic
    wormhole connections
    """
    ctx.obj = _Config(
        relay_url=WELL_KNOWN_MAILBOXES.get(mailbox, mailbox),
        use_tor=bool(ip_privacy),
        debug_file=debug,
        # these will be empty; client must activate ports by sending
        # messages to allow listening (or connecting)
        listen_policy = LocalhostTcpPortsListenPolicy([]),
        connect_policy = LocalhostTcpPortsConnectPolicy([]),
    )
    def run(reactor):
        return ensureDeferred(
            forward(
                reactor,
                ctx.obj,
            )
        )
    return react(run)


@click.option(
    "--debug-messages",
    default=None,
    type=click.File(mode="w", encoding="utf8"),
    help="Save all input/output messages to a file",
)
@click.option(
    "--ip-privacy/--clearnet",
    default=False,
    help="Enable operation over Tor (default is public Internet)",
)
@click.option(
    "--mailbox",
    default=PUBLIC_MAILBOX_URL,
    help='URL for the mailbox server to use (or "default" or "winden" to use well-known servers)',
    metavar="URL or NAME",
)
@click.option(
    "--debug",
    default=None,
    help="Output wormhole state-machine transitions to the given file",
    type=click.File("w", encoding="utf8"),
)
@click.option(
    "--local", "-L",
    multiple=True,
    help=(
        "Listen locally, connect remotely (accepted multiple times)."
        "Unless otherwise specified, (local) bind and (remote) connect addresses are localhost."
        'For example "127.0.0.1:1234:127.0.0.1:22" is the same as "1234:22" effectively.'
    ),
    metavar="[bind-address:]listen-port[:remote-address][:connect-port]",
)
@click.option(
    "--remote", "-R",
    multiple=True,
    help=(
        "Listen remotely, connect locally (accepted multiple times)"
        "Unless otherwise specified, the (remote) bind and (local) connect addresses are localhost."
        'For example "127.0.0.1:1234:127.0.0.1:22" is the same as "1234:22" effectively.'
        ),
    metavar="[remote-bind-address:]listen-port[:local-connect-address][:local-connect-port]",
)
@click.option(
    "--allow-listen",
    multiple=True,
    help=(
        "Accept a connection to this local port. Accepted multiple times."
        "Note that local listeners added via --local are already allowed and do not need this option."
        'If no interface is specified, "localhost" is assumed.'
    ),
    metavar="[interface:]listen-port",
)
@click.option(
    "--allow-connect",
    multiple=True,
    help=(
        "Accept a connection to this local port. Accepted multiple times"
        'If no address is specified, "localhost" is assumed.'
    ),
    metavar="[address:]connect-port",
)
@click.option(
    "--code-length",
    default=2,
    help="Length of the Wormhole code (if we allocate one)",
)
@click.option(
    "--readme", "-r",
    help="Display the full project README",
    is_flag=True,
)
@click.option(
    "--interactive", "-i",
    help="Run in interactive mode, a human-friendly fowld",
    is_flag=True,
)
@click.argument("code", required=False)
@click.command()
def fowl(ip_privacy, mailbox, debug, allow_listen, allow_connect, local, remote, code_length, code, readme, interactive, debug_messages):
    """
    Forward Over Wormhole, Locally

    Bi-directional streaming data over secure and durable Dilated
    magic-wormhole connections.

    This frontend is meant for humans -- if you want machine-parsable
    data and commands, use fowld (or 'python -m fowl')

    This will create a new session (allocating a fresh code) by
    default. To join an existing session (e.g. you've been given a
    code) add the code as an (optional) argument on the command-line.
    """
    if readme:
        display_readme()
        return

    local_commands = [
        _specifier_to_tuples(cmd)
        for cmd in local
    ]
    remote_commands = [
        _specifier_to_tuples(cmd)
        for cmd in remote
    ]

    def to_local(local_interface, local_port, remote_address, remote_port):
        return LocalListener(
            f"tcp:{local_port}:interface={local_interface}",
            f"tcp:{remote_address}:{remote_port}",
        )

    def to_remote(local_interface, local_port, remote_address, remote_port):
        return RemoteListener(
            f"tcp:{local_port}:interface={local_interface}",
            f"tcp:{remote_address}:{remote_port}",
        )

    def to_listen_policy(local_interface, local_port, remote_address, remote_port):
        return local_port

    def to_connect_policy(local_interface, local_port, remote_address, remote_port):
        return remote_port

    def to_iface_port(allowed):
        if ':' in allowed:
            iface, port = allowed.split(':', 1)
            return iface, _to_port(port)
        return "localhost", _to_port(allowed)

    def to_local_port(allowed):
        if ':' in allowed:
            iface, port = allowed.split(':', 1)
            if iface != "localhost":
                raise ValueError(f"Non-local interface: {iface}")
            return _to_port(port)
        return _to_port(allowed)

    def is_local(local_interface, local_port, remote_address, remote_port):
        return is_localhost(local_interface)

    def is_local_connect(local_interface, local_port, remote_address, remote_port):
        return is_localhost(remote_address)

    if any(not is_local(*cmd) for cmd in local_commands) or \
       any(not is_localhost(to_iface_port(allowed)[0]) for allowed in allow_listen):
        listen_policy = ArbitraryInterfaceTcpPortsListenPolicy(
            [(iface, port) for iface, port, _, _ in local_commands] + \
            [to_iface_port(allowed) for allowed in allow_listen]
        )
    else:
        listen_policy = LocalhostTcpPortsListenPolicy(
            [to_listen_policy(*cmd) for cmd in local_commands] +
            [to_local_port(port) for port in allow_listen]
        )

    if any(not is_local_connect(*cmd) for cmd in remote_commands) or \
       any(not is_localhost(to_iface_port(allowed)[0]) for allowed in allow_connect):
        # yes, this says "to_iface_port()" below but they both look
        # the same currently: "192.168.1.2:4321" for example
        connect_policy = ArbitraryAddressTcpConnectPolicy(
            [(addr, port) for _, _, addr, port in remote_commands] + \
            [to_iface_port(allowed) for allowed in allow_connect]
        )
    else:
        connect_policy = LocalhostTcpPortsConnectPolicy(
            [to_connect_policy(*cmd) for cmd in remote_commands] +
            [to_local_port(port) for port in allow_connect]
        )

    cfg = _Config(
        relay_url=WELL_KNOWN_MAILBOXES.get(mailbox, mailbox),
        use_tor=bool(ip_privacy),
        debug_file=debug,
        code=code,
        code_length=code_length,
        commands=[
            to_local(*t)
            for t in local_commands
        ] + [
            to_remote(*t)
            for t in remote_commands
        ],
        listen_policy=listen_policy,
        connect_policy=connect_policy,
        output_debug_messages=debug_messages,
    )

    if interactive:
        return tui(cfg)

    def run(reactor):
        return ensureDeferred(frontend_accept_or_invite(reactor, cfg))
    return react(run)


def _to_port(arg):
    arg = int(arg)
    if arg < 1 or arg >= 65536:
        raise click.UsageError(
            "Ports must be an integer from 1 to 65535"
        )
    return arg


# XXX FIXME use an @frozen attr, not tuple for returns
def _specifier_to_tuples(cmd):
    """
    Parse a local or remote listen/connect specifiers.

    This always returns a 4-tuple of:
      - listen interface
      - listen port
      - connect address
      - connect port

    TODO: tests, and IPv6
    """
    if '[' in cmd or ']' in cmd:
        raise RuntimeError("Have not considered IPv6 parsing yet")

    colons = cmd.count(':')
    if colons > 3:
        raise ValueError(
            f"Too many colons: {colons} > 3"
        )
    if colons == 3:
        # everything is specified
        listen_interface, listen_port, connect_address, connect_port = cmd.split(':')
        listen_port = _to_port(listen_port)
        connect_port = _to_port(connect_port)
    elif colons == 2:
        # one of the interface / address is specified, but we're not
        # sure which yet
        a, b, c = cmd.split(':')
        try:
            # maybe the first thing is a port
            listen_port = _to_port(a)
            listen_interface = "localhost"
            connect_address = b
            connect_port = _to_port(c)
        except ValueError:
            # no, the first thing is a string, so the connect address
            # must be missing
            listen_interface = a
            listen_port = _to_port(b)
            connect_address = "localhost"
            connect_port = _to_port(c)
    elif colons == 1:
        # we only have one split. this could be "interface:port" or "port:port"
        a, b = cmd.split(':')
        try:
            listen_port = _to_port(a)
            listen_interface = "localhost"
            try:
                # the second thing could be a connect address or a
                # port
                connect_port = _to_port(b)
                connect_address = "localhost"
            except ValueError:
                connect_address = b
                connect_port = listen_port
        except ValueError:
            # okay, first thing isn't a port so it's the listen interface
            listen_interface = a
            listen_port = connect_port = _to_port(b)
            connect_address = "localhost"
    else:
        # no colons, it's a port and we're "symmetric"
        listen_port = connect_port = _to_port(cmd)
        listen_interface = "localhost"
        connect_address = "localhost"

    # XXX ipv6?
    return (
        listen_interface, listen_port,
        connect_address, connect_port,
    )


def tui(cfg):
    """
    Run an interactive text user-interface (TUI)

    Allows one to use a human-readable version of the controller
    protocol directly to set up listeners, monitor streams, etc
    """

    def run(reactor):
        return ensureDeferred(frontend_tui(reactor, cfg))
    return react(run)


def display_readme():
    """
    Display the project README
    """
    readme = pkg_resources.resource_string('fowl', '../../README.rst')
    # uhm, docutils documentation is confusing as all hell and no good
    # examples of "convert this rST string to anything else" .. :/ but
    # we should "render" it to text
    click.echo_via_pager(readme.decode('utf8'))


if __name__ == "__main__":
    try:
        import coverage
        coverage.process_startup()
    except ImportError:
        pass
    fowl()



import json

from hypothesis.strategies import one_of, integers, lists, sampled_from, builds, text
from hypothesis import given


from fowl._proto import parse_fowld_command, fowld_command_to_json


def command_messages():
    from fowl import messages
    return [
        (cls, command_class_to_arg_generators(cls))
        for cls in [getattr(messages, nm) for nm in dir(messages)]
        if type(cls) is type and issubclass(cls, messages.FowlCommandMessage) and cls !=  messages.FowlCommandMessage
    ]


def command_class_to_arg_generators(cls):
    from fowl import messages
    return {
        messages.AllocateCode: {
            "length": integers(min_value=1, max_value=32),
        },
        messages.SetCode: {
            "code": text(),
        },
        messages.BytesIn: {
            "id": integers(),
            "bytes": integers(min_value=1),
        },
        messages.BytesOut: {
            "id": integers(),
            "bytes": integers(min_value=1),
        },
        messages.DangerDisablePermissionCheck: {
        },
        messages.LocalListener: {
            "listen": local_server_endpoints(),
            "connect": local_client_endpoints(),
        },
        messages.RemoteListener: {
            "listen": local_server_endpoints(),
            "connect": local_client_endpoints(),
        },
        messages.GrantPermission: {
            "listen": port_lists(),
            "connect": port_lists(),
        },
        messages.Ping: {
            "ping_id": text(),  # should really be "base16-encoded 4-bytes of binary"
        },
    }[cls]


def ports():
    return integers(min_value=1, max_value=65535)


def port_lists():
    return lists(ports())


def local_server_endpoints():
    return sampled_from([
        "tcp:1234:interface=localhost",
    ])


def local_client_endpoints():
    return sampled_from([
        "tcp:localhost:1234",
    ])


all_commands = {
    k: kwargs
    for k, kwargs in command_messages()
}


def commands():
    return one_of([
        builds(k, **kwargs)
        for k, kwargs in all_commands.items()
    ])



@given(commands())
def test_roundtrip(og_cmd):
    """
    Let Hypothesis play with a bunch of round-trip tests for command
    serialization
    """
    parsed_cmd = parse_fowld_command(json.dumps(fowld_command_to_json(og_cmd)))
    assert parsed_cmd == og_cmd, "Command mismatch"




import pytest_twisted

#from twisted.internet.protocol import ProtocolBase
from zope.interface import implementer
from twisted.internet.interfaces import IProcessProtocol
from twisted.internet.protocol import ProcessProtocol
from twisted.internet.defer import DeferredList
from twisted.internet.endpoints import serverFromString, clientFromString
from hypothesis.strategies import integers, sampled_from, one_of, ip_addresses
from hypothesis import given
import click
import sys
import os
import signal
from fowl.observer import When, Framer
from fowl.test.util import ServerFactory, ClientFactory
from fowl.cli import _to_port, _specifier_to_tuples


@implementer(IProcessProtocol)
class CollectStreams(ProcessProtocol):

    def __init__(self, reactor):
        self._reactor = reactor
        self._streams = {
            1: b"",
            2: b"",
        }
        self._done = When()
        self._lines = Framer(reactor)

    def when_done(self):
        return self._done.when_triggered()

    def next_line(self):
        return self._lines.next_message()

    def childDataReceived(self, fd, data):
        self._streams[fd] += data
        if fd == 1:
            self._lines.data_received(data)

    def processExited(self, reason):
        pass  #reason == Failure

    def processEnded(self, reason):
        self._done.trigger(self._reactor, reason)


# maybe Hypothesis better, via strategies.binary() ?
@pytest_twisted.ensureDeferred
async def test_happy_path(reactor, request, mailbox):
    """
    One side invites, other accepts.

    Some commands are executed.
    Improvement: let Hypothesis make up commands, order, etc (how to assert?)
    """

    # NOTE: if debugging this, might want to print output in
    # CollectStreams -- perhaps we need a better solution to
    # "immediately and obviously" print stack-traces there?

    print("Starting invite side", os.environ.get("COVERAGE_PROCESS_STARTUP", "no startup"))

    invite_proto = CollectStreams(reactor)
    invite = reactor.spawnProcess(
        invite_proto,
        sys.executable,
        [
            "python", "-u", "-m", "fowl.cli",
            "--mailbox", mailbox.url,
            # redundant "--allow-connect", "2121",
            "--remote", "2222:2121",
        ],
        env=os.environ,
    )
    request.addfinalizer(lambda:invite.signalProcess(signal.SIGKILL))

    line = await invite_proto.next_line()
    assert line == "Connected."

    line = await invite_proto.next_line()
    assert line.startswith("Secret code: "), "Expected secret code"
    code = line.split(":")[1].strip()

    print(f"Detected code: {code}")

    accept_proto = CollectStreams(reactor)
    accept = reactor.spawnProcess(
        accept_proto,
        sys.executable,
        [
            "python", "-u", "-m", "fowl.cli",
            "--mailbox", mailbox.url,
            "--allow-listen", "2222",
            code,
        ],
        env=os.environ,
    )
    request.addfinalizer(lambda:accept.signalProcess(signal.SIGKILL))

    print("Starting accept side")

    while True:
        result, who = await DeferredList(
            [invite_proto.next_line(), accept_proto.next_line()],
            fireOnOneCallback=True,
        )
        who = "ACC" if who == 1 else "INV"
        print(f"   {who}: {result.strip()}")
        if "Listening:" in result:
            print("  one side is listening")
            break
        if "failed to listen" in result:
            assert False, '"failed to listen" detected in output'

    # now that they are connected, and one side is listening -- we can
    # ourselves listen on the "connect" port and connect on the
    # "listen" port -- that is, listen on 2121 (where there is no
    # listener) and connect on 2222 (where this test is listening)

    listener = ServerFactory(reactor)
    await serverFromString(reactor, "tcp:2121:interface=localhost").listen(listener)  # returns server_port

    client = clientFromString(reactor, "tcp:localhost:2222")
    client_proto = await client.connect(ClientFactory(reactor))
    server = await listener.next_client()

    datasize = 1234
    data = os.urandom(datasize)

    client_proto.send(data)
    msg = await server.next_message(len(data))
    assert msg == data, "Incorrect data transfer"

    print("done")


@given(integers(min_value=1, max_value=65535))
def test_helper_to_port(port):
    assert(_to_port(port) == port)
    assert(_to_port(str(port)) == port)


@given(one_of(integers(max_value=0), integers(min_value=65536)))
def test_helper_to_port_invalid(port):
    try:
        _to_port(port)
        assert False, "Should raise exception"
    except click.UsageError:
        pass


@given(integers(min_value=1, max_value=65535))
def test_specifiers_one_port(port):
    cmd = f"{port}"
    assert _specifier_to_tuples(cmd) == ("localhost", port, "localhost", port)


@given(
    integers(min_value=1, max_value=65535),
    integers(min_value=1, max_value=65535),
)
def test_specifiers_two_ports(port0, port1):
    cmd = f"{port0}:{port1}"
    assert _specifier_to_tuples(cmd) == ("localhost", port0, "localhost", port1)


@given(
    integers(min_value=1, max_value=65535),
    integers(min_value=1, max_value=65535),
    ip_addresses(v=4),  # do not support IPv6 yet
)
def test_specifiers_two_ports_one_ip(port0, port1, ip):
    if ip.version == 4:
        cmd = f"{ip}:{port0}:{port1}"
    else:
        cmd = f"[{ip}]:{port0}:{port1}"
    assert _specifier_to_tuples(cmd) == (str(ip), port0, "localhost", port1)


@given(
    integers(min_value=1, max_value=65535),
    integers(min_value=1, max_value=65535),
    ip_addresses(v=4),  # do not support IPv6 yet
    ip_addresses(v=4),  # do not support IPv6 yet
)
def test_specifiers_two_ports_two_ips(port0, port1, ip0, ip1):
    cmd = f"{ip0}:{port0}:{ip1}:{port1}"
    assert _specifier_to_tuples(cmd) == (str(ip0), port0, str(ip1), port1)


@given(
    integers(min_value=1, max_value=65535),
    integers(min_value=1, max_value=65535),
    ip_addresses(v=6),
    ip_addresses(v=6),
    sampled_from([True, False]),
)
def test_specifiers_unsupported_v6(port0, port1, ip0, ip1, wrap):
    if wrap:
        cmd = f"[{ip0}]:{port0}:[{ip1}]:{port1}"
    else:
        cmd = f"{ip0}:{port0}:{ip1}:{port1}"
    try:
        assert _specifier_to_tuples(cmd) == (str(ip0), port0, str(ip1), port1)
    except RuntimeError:
        pass
    except ValueError:
        pass



import binascii
import automat

from .messages import (
    Welcome,
    CodeAllocated,
    PeerConnected,
    SendMessageToPeer,
    GotMessageFromPeer,
    WormholeClosed,
    PleaseCloseWormhole,
)


class FowlDaemon:
    """
    This is the core 'fowld' protocol.

    This class MUST NOT do any I/O or any asynchronous things (no
    "async def" or Deferred-returning methods).

    There are two main pieces of I/O that we are concerned with here:
    - stdin / stdout, line-based JSON messages
    - wormhole interactions

    Most users of fowld will see the stdin/stdout message (only), as
    that is how we communicate with the outside world.

    Internally (e.g. "tui", "fowl invite", tests) we are concerned
    with both kinds of interactions. Especially for tests, sometimes
    we want to "not actually interact" with the wormhole (but instead
    confirm what this class _wants_ to do).

    Follwing https://sans-io.readthedocs.io/how-to-sans-io.html and
    related ideas and posts, instances of this class interact SOLELY
    through "input" methods, or FowlCommandMessage (via the "input"
    method "command()").

    The `message_handler` you pass here will get instances of
    `FowlOutputMessage` and is expected to "do the I/O" as
    appropriate; in the "normal" case this means either dumping a JSON
    line to stdout or calling a wormhole method (causing I/O to the
    peer or mailbox server).

    On the "input" side, JSON on stdin becomes `FowlCommandMessage`
    instances, which are translated into the correct input method to
    call. In the "normal" case this means parsing incoming lines with
    `parse_fowld_command` and translating that to an appropriate call
    to an `@m.input()` decorated method (e.g. got_welcome(),
    code_allocated(), etc ..)

    XXX: would it be more clear to just have a .command() and de-multiplex in here?
    """
    m = automat.MethodicalMachine()
    set_trace = m._setTrace

    def __init__(self, config, message_handler, command_handler):
        self._config = config
        self._messages = []  # pending plaintext messages to peer
        self._verifier = None
        self._versions = None
        self._message_out = message_handler
        self._command_out = command_handler

    def _emit_message(self, msg):
        """
        Internal helper to pass a message to our external message handler
        (and do something useful on error)
        """
        try:
            self._message_out(msg)
        except Exception as e:
            print(f"Error in user code sending a message: {e}")
            print(type(e), e)

    def _emit_command(self, msg):
        """
        Internal helper to pass a command up to our IO handler
        """
        self._command_out(msg)

    @m.state(initial=True)
    def waiting_code(self):
        """
        """

    @m.state()
    def waiting_peer(self):
        """
        Do not yet have a peer
        """

    @m.state()
    def verifying_version(self):
        """
        Does our peer have compatible versions?
        """

    @m.state()
    def connected(self):
        """
        Normal processing, our peer is connected
        """

    @m.state()
    def closing(self):
        """
        We have asked to close the wormhole, wait until it is
        closed
        """

    @m.state()
    def closed(self):
        """
        Nothing more to accomplish, the wormhole is closed
        """

    @m.input()
    def got_welcome(self, hello):
        """
        We have received a 'Welcome' message from the server
        """

    @m.input()
    def code_allocated(self, code):
        """
        We have a wormhole code (either because it was "set" or because we
        allcoated a new one).
        """
        # this happens either because we "set-code" or because we
        # asked for a new code (i.e. "invite" or "accept") from the
        # OUTSIDE (i.e. controlling function)

    @m.input()
    def peer_connected(self, verifier: bytes, peer_features: dict):
        """
        We have a peer

        :param verifier: a tagged hash of the peers symmetric
            key. This should match what our peer sees (users can
            verify out-of-band for extra security)

        :param peer_features: arbitrary JSON-able data from the peer,
            intended to be used for protocol and other negotiation. A
            one-time, at-startup pre-communication mechanism (definitely
            before any other messages). Also serves as key-confirmation.
        """

    @m.input()
    def version_ok(self, verifier, peer_features):
        """
        Our peer version is compatible with ours
        """

    @m.input()
    def version_incompatible(self, verifier, peer_features):
        """
        We cannot speak with this peer
        """

    @m.input()
    def got_message(self, plaintext):
        """
        We received a message from our peer
        """

    @m.input()
    def send_message(self, plaintext):
        """
        Pass a new message to our peer
        """

    @m.input()
    def shutdown(self, result):
        """
        The wormhole has closed
        """

    @m.output()
    def emit_code_allocated(self, code):
        self._emit_message(CodeAllocated(code))

    @m.output()
    def emit_peer_connected(self, verifier, peer_features):
        """
        """
        self._emit_message(
            PeerConnected(
                binascii.hexlify(verifier).decode("utf8"),
                peer_features,
            )
        )

    @m.output()
    def emit_send_message(self, plaintext):
        self._emit_message(
            SendMessageToPeer(
                plaintext,
            )
        )

    @m.output()
    def emit_got_message(self, plaintext):
        self._emit_message(
            GotMessageFromPeer(
                plaintext,
            )
        )

    @m.output()
    def emit_welcome(self, hello):
        self._emit_message(
            Welcome(self._config.relay_url, hello)
        )

    @m.output()
    def verify_version(self, verifier, peer_features):
        """
        Check that our peer supports the right features
        """
        features = peer_features.get("fowl", {}).get("features", [])
        # note: if we add a feature we will want to do an intersection
        # or something and then trigger different behavior depending
        # what our peer supports .. for now there's only one thing,
        # and it MUST be supported
        from ._proto import SUPPORTED_FEATURES  # FIXME
        if set(features) == set(SUPPORTED_FEATURES):
            self.version_ok(verifier, peer_features)
        else:
            self.version_incompatible(verifier, peer_features)

    @m.output()
    def queue_message(self, plaintext):
        self._message_queue.append(plaintext)

    @m.output()
    def send_queued_messages(self):
        to_send, self._messages = self._messages, []
        for m in to_send:
            self.emit_send_message(m)

    @m.output()
    def emit_shutdown(self, result):
        self._emit_message(
            WormholeClosed(result)
        )

    @m.output()
    def emit_close_wormhole(self):
        self._emit_command(
            PleaseCloseWormhole("versions are incompatible") # XXX hardcoded bad
        )

    waiting_code.upon(
        code_allocated,
        enter=waiting_peer,
        outputs=[emit_code_allocated],
    )
    waiting_code.upon(
        send_message,
        enter=waiting_code,
        outputs=[queue_message]
    )
    waiting_code.upon(
        got_welcome,
        enter=waiting_code,
        outputs=[emit_welcome]
    )
    waiting_code.upon(
        shutdown,
        enter=closed,
        outputs=[emit_shutdown]
    )

    waiting_peer.upon(
        send_message,
        enter=waiting_peer,
        outputs=[queue_message]
    )
    waiting_peer.upon(
        got_welcome,
        enter=waiting_peer,
        outputs=[emit_welcome]
    )
    waiting_peer.upon(
        got_message,
        enter=waiting_peer,
        outputs=[emit_got_message]
    )
    waiting_peer.upon(
        peer_connected,
        enter=verifying_version,
        outputs=[verify_version]
    )
    verifying_version.upon(
        version_ok,
        enter=connected,
        outputs=[emit_peer_connected, send_queued_messages],
    )
    verifying_version.upon(
        version_incompatible,
        enter=closing,
        outputs=[emit_close_wormhole],
    )
    waiting_peer.upon(
        shutdown,
        enter=closed,
        outputs=[emit_shutdown]
    )

    connected.upon(
        send_message,
        enter=connected,
        outputs=[emit_send_message]
    )
    connected.upon(
        got_message,
        enter=connected,
        outputs=[emit_got_message]
    )
    connected.upon(
        shutdown,
        enter=closed,
        outputs=[emit_shutdown]
    )

    closing.upon(
        shutdown,
        enter=closed,
        outputs=[emit_shutdown]
    )
    # XXX there's no notification to go from "connected" to
    # "waiting_peer" -- because Dilation will silently "do the right
    # thing" (so we don't need to). But it would be nice to tell the
    # user if we're between "generations" or whatever



from attr import define, Factory
from twisted.internet.defer import Deferred, succeed


_UNSET = object()


class Framer:
    """
    Takes a stream of bytes and produces 'messages' from them,
    triggering an underlying Next()-style observable.

    Only produces a single message per reactor 'tick'. If there is
    'already' a subsequent message availble, we use a callLater(0,
    ...) to produce it and check further.

    (This is necessary so that something doing an async iteration has
    a chance to "do other work" .. e.g. call .next_message() again
    ... before any more messages are delivered XXX demo of this problem?)
    """

    def __init__(self, reactor): ## just does LineReceiver for now, find_a_frame):
        self._reactor = reactor
        self._data = b""
        self._next = Next()

    def next_message(self):
        """
        :return Awaitable: a new Deferred that fires when a complete,
        as-yet undelivered message has arrived.
        """
        return self._next.next_item()

    def data_received(self, data):
        self._data += data
        self._maybe_deliver_messages()

    def _find_frame(self, data):
        """
        hard-coded to LineReceiver, could make more general

        :returns: 2-tuple of the message and remaining data. if
            "message" is None, there was no complete message yet and
            all data is returned
        """
        if b"\n" in data:
            return data.split(b"\n", 1)
        return (None, data)

    def _maybe_deliver_messages(self):
        msg, self._data = self._find_frame(self._data)
        if msg is not None:
            self._next.trigger(self._reactor, msg.decode("utf8"))
            if self._find_frame(self._data)[0] is not None:
                self._reactor.callLater(0, self._maybe_deliver_messages)


@define
class Next:
    """
    An observerable event that can by async-ly listened for and
    triggered multiple times.

    Used for implementing the a ``next_thing()`` style of method.
    """

    _awaiters: list = Factory(list)
    _unheard_result: object = _UNSET

    def next_item(self):
        """
        :return Awaitable: a new Deferred that fires when this observable
            triggered. This will always be 'in the future' even if we've
            triggered more than zero times already.
        """
        if self._unheard_result is not _UNSET:
            d = succeed(self._unheard_result)
            self._unheard_result = None
        else:
            d = Deferred()
            self._awaiters.append(d)
        return d

    def trigger(self, reactor, result):
        """
        Triggers all current observers and resets them to the empty list.
        """
        listeners, self._awaiters = self._awaiters, []
        if listeners:
            for d in listeners:
                reactor.callLater(0, d.callback, result)
        else:
            self._unheard_result = result


@define
class When:
    """
    An observerable event that can be async-ly listened for and
    triggers exactly once.

    Used for implementing the a ``when_thing()`` style of method.
    """

    _awaiters: list = Factory(list)
    _result: object = _UNSET

    def when_triggered(self):
        """
        :return Awaitable: a new Deferred that fires when this observable
            triggered. This maybe be 'right now' if we already have a result
        """
        if self._result is not _UNSET:
            d = succeed(self._result)
        else:
            d = Deferred()
            self._awaiters.append(d)
        return d

    def trigger(self, reactor, result):
        """
        Triggers all current observers and resets them to the empty list.
        """
        assert self._result is _UNSET, "Can only trigger it once"
        listeners, self._awaiters = self._awaiters, []
        self._result = result
        for d in listeners:
            reactor.callLater(0, d.callback, result)


@define
class Accumulate:
    """
    An observerable event that can by async-ly listened for and
    triggered multiple times, with a per-event 'size' of item to
    collect (as observed via len() calls).

    Used for implementing the a ``next_message(size=123)`` style of method.
    """

    _results: object
    _awaiters: list = Factory(list)

    def next_item(self, reactor, size):
        """
        :return Awaitable: a new Deferred that fires when this observable
            triggered. This will always be 'in the future' even if we've

            triggered more than zero times already.
        """
        d = Deferred()
        self._awaiters.append((size,d))
        self._examine_results(reactor)
        return d

    def some_results(self, reactor, result):
        """
        Append these results. If this gives us enough results to notify
        current listeners, we do.
        """
        self._results += result
        self._examine_results(reactor)

    def _examine_results(self, reactor):
        if not self._awaiters:
            return
        size, d = self._awaiters[0]
        if len(self._results) >= size:
            self._awaiters.pop(0)
            self._results, result = self._results[size:], self._results[:size]
            reactor.callLater(0, d.callback, result)
